\\ Copyright 2018, 2019 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ ...
\\ See examples/BT-rotation.gp for more on this.



\\ Usage: gp <BT-rotation.gp
\\
\\ Some code and notes on binary tree rotations, done in postorder forests;
\\ and Hamiltonian path and cycle through all such by one rotation at a time.

default(recover,0)
default(strictargs,1)
read("../vpar.gp");


\\ J. M. Pallo, "Enumerating, Ranking and Unranking Binary Trees",
\\ The Computer Journal, volume 29, number 2, 1986, pages 171-175.
\\ https://academic.oup.com/comjnl/article/29/2/171/460530
\\
\\ considers a "weight" sequence in a binary tree, where weight of an
\\ external vertex is to take the maximal subtree which it is right-most in,
\\ and count external vertices (including itself) in that subtree.  Pallo's
\\ example is
\\                  *
\\             /       \
\\          *             *
\\       /    \          / \
\\      *      *        *   9
\\     / \    / \      / \
\\    1   2  3   *    6   *
\\              / \      / \
\\             4   5    7   8
\\
\\    vertex  1  2  3  4  5  6  7  8  (9)
\\    weight  1, 2, 1, 1, 5, 1, 1, 3  (9)
\\
\\ For example, vertex 8 is the right-most external of the subtree 6 through
\\ 8, which is 3 external vertices.  Final vertex 9 is right-most of
\\ everything and so would be weight 9, but last vertex is always whole tree
\\ so is omitted.
\\
\\ With an in-order labelling of the internal vertices, the weight is the
\\ number of internal vertices in the subtree comprising a vertex and its
\\ left child subtree (if any).  This follows because each right-most
\\ external label goes to the parent of the top of its subtree, except the
\\ very last external which is ignored.  Pallo's example again,
\\
\\                  5                        5         8
\\             /       \                   / | \      / \
\\          2             8               2  3  4    6   7
\\       /    \          / \              |
\\      1      3        6   e             1
\\     / \    / \      / \
\\    e   e  e   4    e   7            "natural" correspondence,
\\              / \      / \              postorder labelling
\\             e   e    e   e
\\
\\ In the natural correspondence to ordered trees (see vpar.gp initial
\\ comments under "Binary Trees"), in-order labelling of the binary tree
\\ corresponds to post-order labelling of the ordered forest.  The binary
\\ left subtree is all the children of a forest vertex.  Hence the weights
\\ sequence is vpar_subtree_sizes_vector() of post-order forest.

vpar=[2,5,5,5,0, 8,8,0];     \\ Pallo's example, as postorder forest
vpar_validate(vpar);
vpar_is_postorder(vpar)  || error();
vpar_subtree_sizes_vector(vpar) == [1,2, 1,1,5, 1,1,3]  || error(); \\ weights


\\ Pallo gives conditions for when a weights vector is valid, which means
\\ here when a postorder subtree sizes vector.  Vertex v has as its subtree
\\ vec[v] many vertices, including itself, at and preceding v.  Vertices in
\\ that range v-vec[v]+1 .. v, inclusive, must then have their subtrees
\\ entirely within that low limit, so > v-vec[v].
\\
\\ is_postorder_sizes_vector() below is a simple implementation of that
\\ condition.  It can be done faster by noting that the limit on a child c
\\ only needs to be checked against its parent size, since parent limit is
\\ >= grandparent limit, in turn >= great grandparent limit, etc.
\\ vpar_is_postorder_sizes_vector() in vpar-wip.gp (work-in-progress, likely
\\ to change wildly, could disappear) does that.

\\ vec is a vector of integers.  Return 1 if vec is a valid weights vector,
\\ which means subtree sizes vector of a postorder labelled forest.
is_postorder_sizes_vector(vec) =
{
  for(i=1,#vec,
     1 <= vec[i] && vec[i] <= i || return(0);
     for(j=i-vec[i]+1, i-1,
        i-vec[i] <= j-vec[j] || return(0)));
  1;
}


\\-----------------------------------------------------------------------------
\\ Binary Tree Rotation
\\
\\ A "rotation" in a binary tree is a change of an edge between vertices x,y
\\ from leftwards to rightwards, or vice-versa, by rearranging the
\\ parent/child relations.
\\
\\     x,y left                right rotation               x,y right
\\       edge                   ---------->                   edge
\\                   y                               x
\\                /    \                           /   \
\\              x      subtree-C            subtree-A   y
\\            /   \                                   /   \
\\     subtree-A  subtree-B                   subtree-B   subtree-C
\\
\\                              <----------
\\                             left rotation
\\
\\ This is is commonly used for "re-balancing" a binary search tree or
\\ similar.  The order of subtrees A,B,C is unchanged but the size under x
\\ increases and y decreases, or vice versa.
\\
\\ Pallo shows that a rotation changes one entry in the weights vector, and
\\ that the weights vectors form a lattice (the algebra kind of lattice)
\\ under a term-wise partial ordering.
\\
\\ In terms of ordered forests, x,y left edge is x first child of y.  The
\\ following diagram shows the rotation.  Siblings as indicated "..." and
\\ the "further-B,C" are zero or more further vertices and possible subtrees
\\ under them.
\\
\\                             right rotation
\\                              ---------->
\\         y ... further-C                       x ... y ... further-C
\\         |                                     |     |
\\         x ... further-B                 subtree-A  further-B
\\         |
\\     subtree-A
\\                              <----------
\\                             left rotation
\\
\\ Right rotation moves a first child x up to be preceding sibling of its
\\ parent y.  Or conversely left rotation is x preceding sibling of y
\\ dropped down to first child of y.  Right rotation decreases the subtree
\\ size of y, or left rotation increases the subtree size of y.  In
\\ postorder labelling this is a change of x parent, with no renumbering.
\\ In all cases the subtree of x is unchanged (all of A), as are all other
\\ subtrees above and below in the forest.


\\-----------------------------------------------------------------------------
\\ Hamiltonian Path
\\
\\ A "binary tree rotations" graph can be made where each vertex represents
\\ a binary tree, and edges are between those differing by one rotation.
\\
\\ Pallo (figure 1) draws that for binary trees N=4 internal vertices,
\\ arranged in rotation steps.  A 3D drawing using the weights vector
\\ entries as coordinates is also possible, and makes an attractive
\\ rectangle showing the lattice nature.  Higher N as 4D etc projection
\\ starts to have too many lines crossing to draw by vectors.
\\
\\ Each edge in a binary tree can rotate.  Left edges can rotate right, or
\\ right edges can rotate left.  So the graph is degree N-1 regular.
\\
\\ A Hamiltonian path through the graph is binary trees in a sequence where
\\ each differs from the previous by one rotation.  Such a path exists per
\\
\\     Joan Lucas, "The Rotation Graph of Binary Trees Is Hamiltonian",
\\     Journal of Algorithms, volume 9, 1988, pages 503-535.
\\     CS-TR-021-86
\\
\\ And a simplified approach to the same
\\
\\     Joan Lucas, Dominique Roelants van Baronaigien, Frank Ruskey, "On
\\     Rotations and the Generation of Binary Trees", Journal of Algorithms,
\\     volume 15, 1993, pages 343-366.
\\     http://webhome.cs.uvic.ca/~ruskey/Publications/Rotation/Rotation.html
\\
\\ This latter is by induction.  A Hamiltonian path of n-1 vertex trees gets
\\ a further vertex n at all permitted places.  In the weights vectors this
\\ means take each n-1 vector and append the possible additional n'th
\\ values.  In ordered forests, this is which roots of the n-1 forest are to
\\ be children of new vertex n.  For example,
\\
\\                                  14      new v=14
\\                                /
\\            6      9      13
\\           / \     |    / | \
\\          4   5    8   10 11 12
\\         /|\       |
\\        1 2 3      7
\\
\\ The roots of n-1 are 6,9,13.  Postorder n forests are v=14 having no
\\ children, child 13, children 9,13, or children 6,9,13.
\\
\\ If v=14 has either no children or all children (6,9,13) then the n-1 can
\\ step by one rotation in its Hamiltonian sequence and still remain a valid
\\ forest n.
\\
\\ So the idea is take an n-1, run the n vertex up from none to all
\\ children, step n-1, run n vertex down from all to none, step n-1, etc.
\\
\\     n=4 sizes vectors
\\     Hamiltonian path going up and down the n=3 Hamiltonian path
\\
\\     1,1,1,4 --- 1,1,2,4     1,1,3,4 --- 1,2,3,4      1,2,1,4
\\        |           |           |           |            |
\\     1,1,1,3     1,1,2,3        |           |            |
\\        |           |           |           |            |
\\     1,1,1,2        |           |           |         1,2,1,2
\\        |           |           |           |            |
\\     1,1,1,1     1,1,2,1 --- 1,1,3,1     1,2,3,1 ---  1,2,1,1
\\
\\ Knuth volume 4 section 7.2.1.6 has a description of this too, with
\\ pictures for n above n-1.
\\ http://www-cs-faculty.stanford.edu/~knuth/fasc4a.ps.gz
\\
\\ The code below builds vpar forests this way.  vpar for this is probably
\\ not ideal, due to recalculating vpar_roots() each time.  Finding the
\\ roots is a little easier in postorder than the general case (find next by
\\ chasing up parents from v+1).  Or could keep a list of the roots (which
\\ growing to n would use some of the existing and prepend n).
\\
\\ Building sizes (weights) vectors would be better, because the vertex
\\ immediately preceding the subtree of n has its top at n - sizes[n].  So
\\ add sizes[n-sizes[n]] to sizes[n] to put the next n-1 root underneath n.
\\ vpar_make_postorder_sizes() in vpar-wip.gp (again work-in-progress,
\\ likely to change wildly or disappear) is some code to go from sizes
\\ vector to vpar if the latter is then wanted.
\\
\\ The "down" direction in the code is kept explicitly by Lucas et al for
\\ each position.  The path is iterated by stepping position i up or down
\\ according to its direction, or at its limit reverse and step position i-1
\\ instead.


\\ the sample forest above
vpar=[4,4,4,6,6,0, 8,9,0, 13,13,13,0, 0];
vpar_validate(vpar);
vpar_is_postorder(vpar)  || error();


\\ vpars = grow_Hamiltonian_path(vpars)
\\
\\ vpars is a vector of postorder labelled vpar forests (vpar_is_postorder()),
\\ all of the same number of vertices, and where each differs from its
\\ predecessor by one binary tree rotation.
\\ Return a similar vector but of all the vpar forests of 1 more vertex.
\\
\\ n is the number of vertices in the new vpars being made.
\\ roots[] is the set of roots of the existing vpars[i].
\\ new_vpars is num_roots+1 many new vpars of n vertices.  They have some of
\\ the roots[] in vpars[i] changed to have parent the new vertex n.
\\ new_vpars[1] is none under n.
\\ new_vpars[2] is roots[#roots] under n.
\\ new_vpars[3] is roots[#roots-1 .. #roots] under n.
\\ And so on to new_vpars[#roots+1] has all roots[1 .. #roots] under n.
\\
\\ new_vpars is built as none up to all children of n.  It is reversed
\\ according to "down" so as to alternate none to all or all to none.
\\
grow_Hamiltonian_path(vpars) =
{
  my(n = #vpars[1] + 1,
     down = 0,
     ret=vector(#vpars));  \\ vector of new_vpars vectors
  for(i=1,#vpars,
     my(roots=vpar_roots(vpars[i]),
        new_vpars=vector(#roots+1));
     new_vpars[1] = Vec(vpars[i],n);  \\ extended by 1, n childless
     for(j=2,#new_vpars,
        new_vpars[j] = new_vpars[j-1];         \\ copy prev
        new_vpars[j][roots[#roots-j+2]] = n);  \\ root -> child of n
     if(down,new_vpars=Vecrev(new_vpars));
     ret[i] = new_vpars;
     down = !down);
  concat(ret);
}
make_Hamiltonian_path(n) =
{
  my(vpars = [ [0] ]);  \\ start at n=1 vertices
  for(rep=2,n,
     vpars = grow_Hamiltonian_path(vpars));
  vpars;
}

n = 3;            \\ num vertices in forests
vpars = make_Hamiltonian_path(n);

print("Hamiltonian path, post-order labelled trees (count ",#vpars,")");
printf("%*s   %*s\n", 2*n,"sizes", 3*n,"vpar");
printf("%*s   %*s\n", 2*n,"-----", 3*n,"----");
for(i=1,#vpars, \
   print(vpar_subtree_sizes_vector(vpars[i]),"   "vpars[i]));


\\ bool = is_sizes_one_rotation(s1,s2)
\\ s1 and s2 are weights vectors (postorder subtree sizes vectors).
\\ Return 1 if they differ by one binary tree rotation.
\\ This means they differ in one vector entry, and that difference is
\\ the minimum possible difference (nothing in between is a valid weights
\\ vector).
\\ Suspect the test for minimum difference can be done more efficiently than
\\ checking all in between ...
is_sizes_one_rotation(s1,s2) =
{
  my(diff = s1-s2);
  hammingweight(diff)==1 || return(0);   \\ exactly 1 entry different

  \\ Check that no other sizes vectors between s1 and s2 are be valid.
  \\ (This is not a particular efficient test.)
  my(d = vecsum(diff),  \\ the difference value
     step = -diff/abs(d));   \\ vector +/-1 in the differing entry
  \\ print(s1" "s2" diff="diff" "abs(d)-1" step "step);
  for(i=1,abs(d)-1,
     s1 += step;
     if(is_postorder_sizes_vector(s1), return(0)));
  \\ print("  end "s1"  "s1+step);
  s1+step == s2 || error();
  1;
}
\\ vpars is a vector of postorder labelled vpar forests.
\\ Check that each differs from the next by one binary tree rotation.
check_Hamiltonian_path(vpars) =
{
  \\ One rotation means sizes changing in one entry.
  \\ (And change by the smallest possible amount too, but that's not
  \\ checked here.)
  for(i=2,#vpars,
     is_sizes_one_rotation(vpar_subtree_sizes_vector(vpars[i]),
                           vpar_subtree_sizes_vector(vpars[i-1]))
     || error("oops, not a single rotation"));
}

#vpars == vpar_count_ordered(n,1)  || error();    \\ Catalan many forests
#Set(vpars) == #vpars              || error();    \\ no duplicates
for(i=1,#vpars, vpar_validate(vpars[i]));         \\ all valid
for(i=1,#vpars, vpar_is_postorder(vpars[i]) || error());
check_Hamiltonian_path(vpars);

print();


\\-----------------------------------------------------------------------------
\\ Hamiltonian Cycle
\\
\\ It can be noted that the Hamiltonian path for n=4 shown above is not a
\\ cycle.  A Hamiltonian cycle can be formed as per
\\
\\     Ro-Yu Wu, Hung Lung Wang, "A Simple Proof for the Hamiltonian Property on
\\     Rotation Grpahs of Binary Trees" The 27th Workshop on Combinatorial
\\     Mathematics and Computation Theory
\\
\\ Their approach starts from the Hamiltonian path, and a certain base cycle
\\ existing in it from 111111 to 121111.  Other vertices are "uncovered" and
\\ blocks of them are insinuated into the base cycle by seeing they form
\\ cycles with suitable links back to the base cycle.  If n is even then a
\\ final non-cycle uncovered block gets special treatment (it by zig-zag?).
\\
\\ A slightly more direct way is to take n-1 cycle and grow it to an n cycle
\\ the same way that the path up/down does.  The number of n-1 forests is
\\ Catalan(n-1), and if it is odd then the grow has first and last as
\\ different up and down, so not a cycle.
\\
\\ Catalan(n-1) is odd if and only if n-1 = 2^k-1, so n=2,4,8,16,... (by
\\ working through the powers-of-2 in the factorials, or Deutsch and Sagan).
\\ When n is even it can be noticed that the first two columns 111x and 112x
\\ can be traversed by zig-zag, thereby making a single down instead of an
\\ up-down.  Doing so when C(n-1) is odd balances the ups and downs.
\\ 
\\     n=4 sizes vectors, Hamiltonian cycle by first two colums zig-zag
\\
\\     +--------------------------------------------------------------+
\\     |                                                              |
\\     +-- 1,1,1,4 --- 1,1,2,4     1,1,3,4 --- 1,2,3,4      1,2,1,4 --+
\\                        |           |           |            |
\\         1,1,1,3 --- 1,1,2,3        |           |            |
\\            |                       |           |            |
\\         1,1,1,2                    |           |         1,2,1,2
\\            |                       |           |            |
\\         1,1,1,1 --- 1,1,2,1 --- 1,1,3,1     1,2,3,1 ---  1,2,1,1
\\
\\ The code below uses the grow_Hamiltonian_cycle() to make an up-down, then
\\ the n>=4 power-of-2 case re-orders the initial 2n-2.  A reversal of the
\\ rest keeps 1111 and 1112 as the first two sizes.
\\
\\ The printout is n=4 which is the first needing the zig-zag.  Set instead
\\ n=8 for 1430 forests to see the next.  n=16 is 35 million forests which
\\ is a bit too much to hold in memory.  Some recursive generating or a
\\ perhaps moderately complicated iterator ought to be possible if that much
\\ really wanted.


initial_reorder(vec) =
{
  \\ Going upwards from 1,1.
  \\ Take the initial 1,1.
  \\ Then two 1,2 and 1,3 and two 2,3 and 2,4, and so on two from each
  \\ column.
  \\ vec has the 2,3, 2,4 etc running from 2,3 last in vec, 2,4 second last,
  \\ etc, hence j decrement.

  (#vec-2)%4 == 0  || error();
  my(ret=vector(#vec),
     i=0, j=#vec+1,
     r=0);
  ret[r++]=vec[i++];
  for(t=1,(#vec-2)/4,
     ret[r++]=vec[i++]; ret[r++]=vec[i++];
     ret[r++]=vec[j--]; ret[r++]=vec[j--]);
  ret[r++]=vec[i++];
  ret;
}
{
  my(n=6);
  print("Hamiltonian cycle initial re-ordering (n="n")");
  vec = concat(vector(n,i, 10+i), vector(n-2,i, 20+n+1-i));
  print("up-down  ",vec);
  print("becomes");
  print("zig-zag  ",initial_reorder(vec));
  print();
}

is_pow2(n) = n==(1<<valuation(n,2));

\\ vpars = grow_Hamiltonian_cycle(vpars)
\\
\\ vpars is a vector of postorder vpar forests like grow_Hamiltonian_path()
\\ above, but a cycle not just a path.
\\ vpars[1] is sizes [1,1,1,1,1,1], which means forest of all singletons.
\\ vpars[2] is sizes [1,1,1,1,1,2], which means singletons and final path-2.
\\ The return is likewise, but one bigger forests.
\\
grow_Hamiltonian_cycle(vpars) =
{
  vpars = grow_Hamiltonian_path(vpars);
  my(n=#vpars[1]);
  if(n>=4 && is_pow2(n),
     vpars = concat(initial_reorder(vpars[1 .. 2*n-2]),
                    Vecrev(vpars[2*n-1 .. #vpars])));
  vpars;
}
make_Hamiltonian_cycle(n) =
{
  my(vpars = [ [0] ]);  \\ start at n=1 vertices
  for(rep=2,n,
     vpars = grow_Hamiltonian_cycle(vpars));
  vpars;
}

vpars = make_Hamiltonian_cycle(n);

print("Hamiltonian cycle");
printf("%*s   %*s\n", 2*n,"sizes", 3*n,"vpar");
printf("%*s   %*s\n", 2*n,"-----", 3*n,"----");
for(i=1,#vpars, \
   print(vpar_subtree_sizes_vector(vpars[i]),"   "vpars[i]));

#vpars == vpar_count_ordered(n,1)  || error();    \\ Catalan many forests
#Set(vpars) == #vpars              || error();    \\ no duplicates
for(i=1,#vpars, vpar_validate(vpars[i]));         \\ all valid
for(i=1,#vpars, vpar_is_postorder(vpars[i]) || error());
check_Hamiltonian_path(concat(vpars,[vpars[1]])); \\ cycling around again

\\-----------------------------------------------------------------------------
