\\ Copyright 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------

\\ Kevin Ryde, September 2021
\\
\\ Usage; gp <a347539.gp
\\
\\ This is some PARI/GP code for the Matula-Goebel codes of rooted trees in
\\ the order of Beyer and Hedetniemi's iteration, and the converse:
\\
\\     A347539   Matula-Goebel number of iteration row n.
\\     A347540   Iteration row number of Matula-Goebel number n.

default(recover,0)
default(strictargs,1)


\\-----------------------------------------------------------------------------
\\ next_levels()
\\
\\ This is similar to the code in A346913.  Here the levels vector is a
\\ parameter and the return is the next levels vector.
\\
\\ Returning a vector means copying so time is linear in the #levels length,
\\ rather than constant time per tree which is the aim of Beyer and
\\ Hedetniemi's iteration.  But the vectors here are small, and
\\ levels_to_primecode() examines the whole vector anyway and so takes time
\\ at least linear in #levels.

\\ "levels" is a vector containing a rooted tree levels sequence which is a
\\ row from A346913.  Return the next levels vector, being the next row of
\\ A346913.  An empty vector levels=[] is accepted too and its next is the
\\ first not-empty (row n=1 of A346913, singleton [1]).
next_levels(levels) =
{
  forstep(p=#levels,3,-1,
     levels[p]>=3 || next;         \\ find p = end-most with a grandparent
     forstep(q=p-1,2,-1,
        levels[q]<levels[p] || next;  \\ find q = parent of p

        \\ change levels[p..#levels] to copies of levels[q .. p-1],
        \\ including partial copy at the end if necessary
        while(p<=#levels,
              levels[p] = levels[q];
              p++; q++);
        return(levels)));

  \\ no vertex with level>=3 exists, so no more of this #levels length,
  \\ step to the first of the next longer length
  [1..#levels+1];
}


\\-----------------------------------------------------------------------------
\\ A347539 = Matula-Goebel codes of Beyer and Hedetniemi's iteration

\\ levels_to_primecode() converts from levels vector to Matula-Goebel
\\ primecode.  It works from the end-most vertex and forms products of
\\ primes in vector pc[].
\\ 
\\ At vertex v in the loop, the level of v is L=levels[v] and the primecode
\\ of the subtree at and below v is pc[L].  This subtree is accumulated into
\\ the parent of v by multiplying pc[L-1] *= prime(pc[L]).  The parent
\\ vertex is not known yet, but it will be level L-1.
\\
\\ Products start as pc[level] = 1 which is the primecode for a childless
\\ vertex (a singleton).  When vertex v has been processed, its pc[L] entry
\\ is cleared back to 1 ready to accumulate possible further subtrees at
\\ depth L earlier in the tree.
\\
\\ The loop stops when v=2 has been accumulated into its parent v=1.
\\ v=1 is the root and  its product pc[1] is the final primecode.
\\
levels_to_primecode(levels) =
{
  my(pc=vector(#levels+1,i,1));
  forstep(v=#levels,2,-1,
     my(L=levels[v]);
     pc[L-1] *= prime(pc[L]);   \\ v accumulate into parent
     pc[L] = 1);                \\ clear back to 1
  pc[1];
}

\\ return a vector which is the first len many terms of A347539
A347539_vector(len) =
{
  my(levels=[]);  \\ empty vector preceding the first actual levels
  vector(len,i,
         levels=next_levels(levels);    \\ step
         levels_to_primecode(levels));  \\ encode
}

{
  print1("A347539 = ");
  my(vec=A347539_vector(18));
  for(i=1,#vec, print1(vec[i],","));
  print("...");
}


\\-----------------------------------------------------------------------------
\\ A347540 = row number in Beyer and Hedetniemi's iteration for a given
\\ Matula-Goebel primecode, so the inverse permuation of A347539.

\\ levels_to_primecode_with_limit() is like levels_to_primecode(), but if
\\ primecode > limit then return symbol 'toobig instead.  Primecode bigger
\\ than limit is checked at each point in the encoding for an early exit.
\\ The limit is not enforced against a singleton levels=[1] because that's
\\ not needed for the way A347540_vector() calls here.
levels_to_primecode_with_limit(levels,limit) =
{
  my(pc=vector(#levels+1,i,1));
  forstep(v=#levels,2,-1,
     my(L=levels[v]);
     pc[L-1] *= prime(pc[L]);
     if(pc[L-1] > limit, return('toobig));
     pc[L] = 1);
  pc[1];
}

\\ return a vector which is the first len many terms of A347540
A347540_vector(len) =
{
  my(ret=vector(len),
     n=1,         \\ n = B+H row number
     levels=[1]); \\ row n=1
  while(len>0,    \\ len = how many primecodes in 1..#ret not yet found
        my(primecode=levels_to_primecode_with_limit(levels,#ret));
        if(primecode!='toobig,
           ret[primecode] = n; len--);
        levels=next_levels(levels); n++);
  ret;
}

{
  print1("A347540 = ");
  my(vec=A347540_vector(18));
  for(i=1,#vec, print1(vec[i],","));
  print("...");
}

\\ check A347539 and A347540 are inverses
{
  my(limit=500,
     n_to_primecode=A347539_vector(limit),
     primecode_to_n=A347540_vector(limit));
  for(i=1,#n_to_primecode,
     if(n_to_primecode[i]<#primecode_to_n,
        primecode_to_n[n_to_primecode[i]] == i  || error()));
  for(i=1,#primecode_to_n,
     if(primecode_to_n[i]<#n_to_primecode,
        n_to_primecode[primecode_to_n[i]] == i  || error()));
}

\\ A347540_vector() iterates through all level sequences until having found
\\ all the desired primecodes 1..len.  This is not efficient.
\\
\\ The main problem is that A005517, the smallest primecode of n vertices,
\\ grows relatively slowly.  For example A005517(1125) = 14 so that finding
\\ primecodes 1..1125 requires iterating through roughly A087803(14) = 53272
\\ trees, of which only 1125 are actually wanted.
\\
\\ levels_to_primecode_with_limit() saves work on the unwanted trees by an
\\ early exit if the maximum desired primecode is exceeded.  This avoids
\\ large primes.  For example, the largest primecode of 13 vertices is
\\ A005517(13) = 997525853 which is a big prime and would need a primes
\\ table about 50 million entries long, even though the target might be
\\ merely primes up to 1125.
\\
\\ The actual operation for A347540 is an "ranking", ie. where in the
\\ iteration does the levels sequence, as found by factoring the primecode,
\\ occur.  Beyer and Hedetniemi note (end of section 6) that both ranking
\\ and unranking the iteration might be quite difficult.
\\
\\ In all pre-order trees, ranking and unranking are usual operations and
\\ are done efficiently using counts from the Catalan triangle.  Only how
\\ many vertices remain and the level of the preceding vertex are needed and
\\ relevant counts are a straightforward recurrence or binomial.  But in the
\\ canonical rooted trees, the whole preceding tree structure is needed
\\ (in general), not just one preceding vertex level.


\\-----------------------------------------------------------------------------
print("end");
