\\ Copyright 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, November 2021

\\ Usage; gp <a348959.gp
\\
\\ a(n) calculates a term of A348959, the childless terminal Wiener index of
\\ Matula-Goebel tree number n.

default(recover,0);
default(strictargs,1);

{
  \\ return 2-element vector [W,t]
  my(W_and_terminals(n) =
       if(n<=1, [0,1],
          my(f=factor(n),wt);
          [ (wt=self()(primepi(p))) + [wt[2]*('total-wt[2]), 0]
           | p<-f[,1] ]   \\ each prime
          * f[,2]));      \\ scalar product for repetitions

  a(n) = my(wt=W_and_terminals(n)); subst(wt[1],'total,wt[2]);
}
{
  print("A348959 childless terminal Wiener:");
  print1("a(n) = ");
  for(n=1,25,
     print1(a(n),","));
  print();
}

\\ This code uses the edge crossings method.  Suppose an edge connects
\\ vertex v down to a child u,
\\
\\               v
\\              / \\       edge u--v crossings
\\             u             t * (total - t)
\\            /|\
\\       t many childless
\\     vertices at and below u
\\
\\ The paths which go through this edge are those from each of the t many
\\ childless vertices under u, up to each of the total-t many childless
\\ vertices on the other side.  Here "total" is the total number of
\\ childless vertices in the tree.  So this edge is crossed t*(total-t)
\\ times and summing over all edges is the lengths of all paths,
\\
\\     Wiener index =  Sum     t[u] * (total - t[u])
\\                   edge u--v
\\
\\ The way the code recurses means "total" is not known until the end, after
\\ all descents are complete.  So "W", which is the sum of edge crossings
\\ so-far, uses symbol 'total in a little polynomial (degree 1) ready to
\\ substitute at the end.  For example 3*('total-3).
\\
\\     W_and_terminals(n)
\\     =
\\     [W,     Wiener index of tree n as a polynomial in 'total
\\      t]     number of terminal vertices within tree n
\\
\\ factor() gives prime powers as p^reps.  Exponent reps means repeats of
\\ the subtree p.  These repetitions are as easy as multiply reps*W for
\\ copies of W in the subtrees, and reps*t*(total-t) for copies of the edge
\\ crossings.  Scalar product [..vector..]*f[,2] is convenient for these
\\ multiples and sum.
\\
\\ There are 3 quantities returned by W_and_terminals(n), being two coeffs
\\ in the polynomial, and t.  The top-level values which a(n) sees are
\\
\\     W = S - 'total*E
\\
\\     E = sum of the depth of each childless vertex, being A196048(n),
\\         the "external path length" of the tree.
\\
\\     S = sum of t[u]^2, so sum of squared number of childless vertices in
\\         the subtree at and below each vertex u, excluding the root vertex.
\\
\\     t = number of childless vertices = A109129(n),
\\         but with the exception that for n=1 here t=1 instead of the way
\\         A109129(1)=0 does not count the root of a singleton.  (This has
\\         no effect anyway, since W=0 for a singleton.)
\\

\\-----------------------------------------------------------------------------
print("end");
