\\ Copyright 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Usage; gp <a750002.gp
\\
\\ Find A750002 = Matula-Goebel numbers of equimatchable trees by a
\\ predicate test.
\\
\\
\\ An equimatchable tree can be identified by working up from the childless
\\ vertices.  At a vertex v, a "state" is calculated based on states of each
\\ child subtree below it.
\\
\\ 5 states suffice and it happens that each state s = 1 to 5 corresponds to
\\ the conditions at the top of a path of s vertices.
\\
\\ In the following table, each size quantity is the difference
\\
\\     smallest maximal matching size - maximum matching size
\\
\\ or "none" when there is no maximal matching at all possible.
\\
\\                size, relative to maximum,
\\                   of smallest maximal
\\     state     with  without_max  without
\\     -----     ----  -----------  -------
\\       1       none       0          0
\\       2         0      none        -1
\\       3         0        0          0   
\\       4         0       -1         -1   
\\       5         0        0         -1   
\\
\\ "with" is relative size of the smallest maximal in which v is matched.
\\ "without_max" is relative size of smallest maximal in which v unmatched.
\\ "without" is relative size of smallest matching with v unmatched and
\\   which is either maximal already or will become maximal if vertex v is
\\   matched to something else later.
\\
\\ The starting state is state=1 for a singleton vertex v.  This states has
\\ with=none since there are no maximal matchings at all where v is matched.
\\
\\
\\ The accepting states are when
\\
\\     with = 0 or none   AND  without_max = 0 or none
\\
\\ This means that any maximals "with" are equal to the maximum size, and
\\ likewise any maximals "without" are equal to the maximum size.
\\
\\     state = 1, 2, 3,    5     accepting
\\     state =          4        not accepting
\\
\\ State 4 is the conditions at the top of a path-4.  It has a maximal
\\ matching (where v is unmatched) of size -1 relative to the maximum
\\ matching, so it is not equimatchable:
\\
\\     v      path 4               v  \    path 4
\\     |      maximal              |  |    maximum
\\     x  \   matching             x  /    matching,
\\     |  |   size = 1             |       size = 2
\\     y  /                        y  \
\\     |                           |  |
\\     z                           z  /
\\
\\ But state 4 is not doomed.  If v in state 4 gets a new singleton child w
\\ then the resulting path-5 has all maximals the same size.  Only as a
\\ final state is 4 no good.
\\
\\
\\ The table[] of transitions below follows by considering combinations of
\\ conditions in the parent and child.  Once a smallest maximal is -2 or
\\ more behind the maximum, it can never reach equality with the maximum
\\ again and the tree is certainly not equimatchable.  table[v,c] = 0 in
\\ that case, and the code can give an early exit "no".
\\
\\ In the factoring, f[i,2] is the number of copies of a child subtree.
\\ Each is a transition.  Though in fact at most 2 copies suffice for the
\\ transitions because 2 copies of any child reaches either dead "0" or the
\\ 3rd and subsequent transitions are no change.


{
  \\ v is the state of a tree and c is the state of a child subtree
  \\ to be added underneath it.
  \\
  \\ table[v,c] = new state at v
  \\ or
  \\ table[v,c] = 0 when the tree now is not equimatchable, and cannot
  \\ become equimatchable, no matter what child subtrees below or what
  \\ structure above
  \\
  \\     c=1  2  3  4  5
  table = [2, 3, 4, 5, 0;   \\ v=1
           2, 0, 2, 0, 0;   \\ v=2
           0, 3, 0, 0, 0;   \\ v=3
           2, 0, 4, 0, 0;   \\ v=4
           0, 0, 0, 0, 0];  \\ v=5
}

equimatchable_state(n) =
{
  my(f=factor(n),
     v=1);  \\ state for a singleton
  for(i=1,matsize(f)[1],
     my(c=equimatchable_state(primepi(f[i,1])));  \\ child subtree
     if(!c, return(0));

     for(j=1,f[i,2],  \\ repetitions of the child
        v=table[v,c];
        if(!v, return(0))));
  v;
}

is(n) = my(state=equimatchable_state(n)); state!=0 && state!=4;

{
  print("A750002 equimatchable tree codes:");
  for(n=1,200,
     if(is(n),
        print1(n", ");
        if(n==49||n==133, print())));
  print("\n");
}
{
  print("A750003 not-equimatchable tree codes:");
  for(n=1,88,
     if(! is(n),
        print1(n", ");
        if(n==36||n==62, print())));
}

print("end");
