\\ Copyright 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

\\-----------------------------------------------------------------------------
\\ Kevin Ryde, October 2022
\\
\\ This is some PARI/GP code to calculate terms of sequences on
\\ Colijn-Plazzotta tree number n,
\\ 
\\     A800001    number of vertices with same child subtrees
\\     A800002    number of vertices with different child subtrees
\\     A800003    number of automorophism orbits
\\ 
\\ Requires PARI/GP 2.13 for sqrtint() remainder.


default(recover,0);
default(strictargs,1);

\\ CP_children(n), for n>=2, returns a vector [x,y] of the child subtrees of n,
\\ x = A002024(n-1) and y = A002260(n-1)
CP_children(n) =
{
  my(r, s=sqrtint((n-1)<<1,&r));
  if(r>s,  r-=s;s++, r+=s);
  [s,r>>1];
}

A800001_num_same_child_subtrees(n) =
{
  if(n==1, 0,
     my(x,y); [x,y] = CP_children(n);
     if(x==y, 1 + 2*self()(x),
                  self()(x) + self()(y)));
}
A800002_num_different_child_subtrees(n) =
{
  if(n==1, 0,
     my(x,y); [x,y] = CP_children(n);
     if(x==y,     2*self()(x),
              1 + self()(x) + self()(y)));
}
A800003_num_orbits(n) =
{
  1 + if(n>1, vecsum(apply(self(),
                           Set(CP_children(n)))));
}

{
  print1("A800001 =");
  for(n=1,15, print1(A800001_num_same_child_subtrees(n),","));
  print("  ...");
}
{
  print1("A800002 =");
  for(n=1,15, print1(A800002_num_different_child_subtrees(n),","));
  print("  ...");
}
{
  print1("A800003 =");
  for(n=1,15, print1(A800003_num_orbits(n),","));
  print("  ...");
}

\\ All 3 functions are recursive traversals of the tree, watching for x==y
\\ or x!=y, and for the orbits counting vertices in just one subtree if
\\ x==y.
\\
\\ Peak recursion depth is modest since the smallest n of height h is
\\ n = A108225(h) which is very large very quickly.
\\
\\ Subtree vectors returned by traverse() are concatenated which ends up
\\ quite a lot of vector copying when the tree is large.  This takes a
\\ measurable time, by comparing returning just a vertex count say.
\\ 
\\ But speed measures roughly the same for listput() to a call-by-reference
\\ destination, or even just storing into a pre-sized vector in a global
\\ variable.


\\-----------------------------------------------------------------------------
print("end");
