\\ Copyright 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(strictargs,1)
default(recover,0);
read("../vpar.gp");


\\ Usage: gp <fewest-maximum-matchings.gp
\\
\\ This is some code counting trees with the fewest maximum matchings per
\\
\\     Clemens Heuberger and Stephan Wagner, "The Number of Maximum
\\     Matchings In a Tree", Discrete Mathematics, volume 311, issue 21,
\\     November 2011, pages 2512-2542.
\\     https://arxiv.org/abs/1011.6554
\\
\\     https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3226351/
\\     full text html
\\
\\ Their theorem 1.1, which is a relatively brief part of the paper,
\\ considers how few maximum matchings a tree of n vertices may have.


\\-----------------------------------------------------------------------------
default(strictargs,0)
default(recover,0)

\\ Clearly the fewest possible is just 1 maximum matching.  This must be a
\\ perfect matching (vpar_has_perfect_matching()), since an unmatched v
\\ vertex could have its matched neighbour p change to match p-v instead.
\\ A perfect matching is possible only for an even number of vertices.
\\
\\ OEIS A005751 is how many free trees of 2n vertices have a perfect
\\ matching.

\\ vec = count_free_perfect_vector(len) returns a vector of length "len"
\\ where vec[n] is the number of free trees of 2*n vertices which have a
\\ perfect matching.
count_free_perfect_vector(len) =
{
  \\ r = A000151 rooted unlabelled perfect matched 
  \\ p = A005750 planted rooted unlabelled perfect matched 
  my(r=vector(len),
     p=vector(len));
  r[1]=1;
  p[1]=1;
  for(n=2,len,
     for(i=1,n-1,
        my(s=sumdiv(n-i,d, d*r[d])); r[n] += r[i]*s; p[n] += p[i]*s);
     r[n] *= 2/(n-1);
     p[n] /= n-1;
     r[n] = sum(i=1,n-1, r[i] * sumdiv(n-i,d, d*r[d])) * 2/(n-1);
     p[n] = sum(i=1,n-1, p[i] * sumdiv(n-i,d, d*r[d]))    /(n-1);
     );
  \\ print(r);
  \\ print(p);

  vector(len,n, (r[n]
                 - sum(m=1,n-1, r[m]*r[n-m])
                 + if(n%2==0, r[n/2], p[(n+1)/2])) / 2);

  my(r=vector(len,i,[1,1]));
  for(n=2,len,
     r[n] = sum(i=1,n-1, r[i] * sumdiv(n-i,d, d*r[d][1])) / (n-1);
     r[n][1]*=2);
  \\ print(r);
  \\ print(p);

  vector(len,n, (r[n][1]
                 - sum(m=1,n-1, r[m][1]*r[n-m][1])
                 + if(n%2==0, r[n/2][1], r[(n+1)/2][2])) / 2);
}

{
  my(vec=count_free_perfect_vector(10));
  print1("free perfect:        ");
  for(n=1,#vec, print1(vec[n],","));
  print("...  (A005751)");
}
{
  my(want=[1,1,2,5,15,49,180,701,2891,12371,54564,246319,1133602]);
  count_free_perfect_vector(#want) == want  || error();
}


\\ {
\\   \\ Compare against an explicit count of the free trees with perfect
\\   \\ matchings.  The number of free trees grows quickly so this is feasible
\\   \\ only for small n.
\\   my(got=vector(8));
\\   for(n=1,#got,
\\      my(vpar=vpar_free_first(2*n), count=0);
\\      until((vpar=vpar_free_next(vpar))=='none,
\\            count += vpar_has_perfect_matching(vpar));
\\      got[n]=count);
\\   got == count_free_perfect_vector(#got)  || error();
\\ }

\\-----------------------------------------------------------------------------
\\ For n odd, there is 1 unmatched vertex and as noted above it can swap
\\ pairings with a neighbour so that there is at least 2 maximum matchings.
\\
\\ Heuberger and Wagner show this fewest 2 maximum matchings is attained for
\\ a perfect matched tree and duplicate one of its leaves.  The unmatched
\\ vertex then has a single neighbour, and swap of pairings is then just 2
\\ ways.
\\
\\ A free tree with one leaf selected to be duplicated is equivalent to
\\ choosing a leaf to become the root.  A root with only one child is called
\\ "planted".

count_unlabelled_planted_perfect_vector(len) =
{
  \\ r = A000151 rooted unlabelled perfect matched 
  \\ p = A005750 planted rooted unlabelled perfect matched 
  my(r=vector(len),
     p=vector(len));
  r[1]=1;
  p[1]=1;
  for(n=2,len,
     for(i=1,n-1,
        my(s=sumdiv(n-i,d, d*r[d])); r[n] += r[i]*s; p[n] += p[i]*s);
     r[n] *= 2/(n-1);
     p[n] /= n-1;
     r[n] = sum(i=1,n-1, r[i] * sumdiv(n-i,d, d*r[d])) * 2/(n-1);
     p[n] = sum(i=1,n-1, p[i] * sumdiv(n-i,d, d*r[d])) / (n-1);
     );
  p;
}

{
  my(vec=count_unlabelled_planted_perfect_vector(10));
  print1("planted perfect:     ");
  for(n=1,#vec, print1(vec[n],","));
  print("...  (A005751)");
}
{
  my(want=[1,1,3,10,39,160,702,3177,14830,70678,342860,1686486]);
  count_unlabelled_planted_perfect_vector(#want) == want  || error();
}
\\ {
\\   \\ Compare against an explicit count of the unlabelled trees with perfect
\\   \\ matchings.  The number of unlabelled trees grows quickly so this is
\\   \\ feasible only for small n.
\\   my(got=vector(7));
\\   for(n=1,#got,
\\      my(vpar=vpar_unlabelled_first(2*n), count=0);
\\      until((vpar=vpar_unlabelled_next(vpar))=='none,
\\            my(root=vpar_root(vpar));
\\            count += (vpar_num_children_of_vertex(vpar,1)==1  \\ root
\\                      && vpar_has_perfect_matching(vpar)));
\\      got[n]=count);
\\   got == count_unlabelled_planted_perfect_vector(#got)  || error();
\\ }

\\-----------------------------------------------------------------------------

\\ vec = count_fewest_maximum_matchings_vector(len) returns a vector of
\\ length "len" where vec[n] is the number of free trees with the fewest
\\ possible maximum matchings (which is 1 for n=1 and even n, or 2 for odd
\\ n>=3).
\\
count_fewest_maximum_matchings_vector(len) =
{
  my(f=count_free_perfect_vector(floor(len/2)),
     p=count_unlabelled_planted_perfect_vector(ceil(len/2)));
  vector(len,n, if(n==1,1, n%2==0,f[n/2], p[(n-1)/2]));
}

{
  my(vec=count_fewest_maximum_matchings_vector(14));
  print1("free fewest maximum: ");
  for(n=1,#vec, print1(vec[n],","));
  print("...");
}
{
  \\ Compare against a search for how many trees of fewest maximum
  \\ matchings.  "fewest" is the number of maximum matchings, and count is
  \\ how many of them seen so far.  A new smaller "fewest" seen resets the
  \\ count.  The number of free trees grows quickly so this is feasible only
  \\ for small n.
  my(got=vector(14));
  for(n=1,#got,
     my(vpar=vpar_free_first(n),
        fewest=2^n+1,  \\ more than anything actual
        count=0);
     until((vpar=vpar_free_next(vpar))=='none,
           my(matchings=vpar_matchnum_and_count(vpar)[2]);
           if(matchings==fewest, count++,
              matchings<fewest, count=1;fewest=matchings));
     got[n]=count);
  got == count_fewest_maximum_matchings_vector(#got)  || error();
}


\\ ; r[i];
\\      r[i][1]  == floor(r[i][1])  || print(r[i][1]);

count_fewest_maximum_matchings_compact(n) =
{
  \\ my(h=n\2, r=vector(h,i,[1,1]));
  \\ for(i=2,h,
  \\    r[i] = sum(j=1,i-1, r[j] * sumdiv(i-j,d, d*r[d][1]))/(i-1);
  \\    r[i][1]*=2);
  \\ if(n%2, r[h][2],
  \\    (r[h][1] + if(h%2==0, r[h/2][1] - r[h/2][1]^2, r[(h+1)/2][2])) / 2
  \\    - sum(i=1,(h-1)\2, r[i][1]*r[h-i][1]));

  \\ my(h=n\2,r=vector(h),p=r,s=r); r[1]=p[1]=1;
  \\ for(i=1,h-1, 
  \\    s[i] = sumdiv(i,d, d*r[d]);
  \\    r[i+1] = sum(j=1,i, r[j]*s[i-j+1]) *2/i;
  \\    p[i+1] = sum(j=1,i, p[j]*s[i-j+1])   /i);
  \\ if(n%2, p[h],
  \\    (r[h] + if(h%2, p[(h+1)/2], r[h/2]-r[h/2]^2)) /2
  \\    - sum(i=1,(h-1)\2, r[i]*r[h-i]));

  if(n<3,1,  my(h=n\2, r=vector(h), s=r); r[1]=1;
    for(i=1,h-1, s[i] = sumdiv(i,d, d*r[d]);
       r[i+1] = sum(j=1,i, r[j]*s[i-j+1]) *2/i);
    my(p=vector(if(n%2,h, h%2,(h+1)/2, 0), i,1));
    for(i=1,#p-1, p[i+1] = sum(j=1,i, p[j]*s[i-j+1]) /i);
    if(n%2, p[h],
       (r[h] + if(h%2, p[#p], r[h/2]-r[h/2]^2) )/2
       - sum(i=1,(h-1)\2, r[i]*r[h-i])));
}
count_fewest_maximum_matchings_compact(2) == 1
count_fewest_maximum_matchings_compact(1) == 1
count_fewest_maximum_matchings_compact(0) == 1

vector(10,n,n*=2; count_fewest_maximum_matchings_compact(n))
count_free_perfect_vector(10)
vector(10,n,n=2*n+1; count_fewest_maximum_matchings_compact(n))
count_unlabelled_planted_perfect_vector(10)
quit

\\xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
v=count_fewest_maximum_matchings_vector(37);
print1("1,");
for(i=1,#v, print1(v[i],","); if(i==21||i==30,print()))

{
  print1(" ");
  for(n=4,7,
     print("n="n);
     my(vpar=vpar_free_first(n),
        want=if(n%2==0,1,2));
     until((vpar=vpar_free_next(vpar))=='none,  \\ trees
           if(vpar_matchnum_and_count(vpar)[2]==want,print(vpar));
           count++);
     print(count);
  );
  quit;
}


\\-----------------------------------------------------------------------------
