# Copyright 2017, 2018, 2019 Kevin Ryde

# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl gentreeg-to-try.pl
#
# This is some rough code, might change wildly, be renamed, etc.
#
# Create GP files /tmp/trees-1.gp, /tmp/trees-2.gp, etc which contain calls
# like
#
#     try([0,1,1]);
#
# The idea is for a GP script to make a try() function and read() desired
# trees files to try something on the free trees of n vertices
#
#     try(vpar) = print(vpar" diameter is "vpar_diameter(vpar));
#     read("/tmp/trees-6.gp");
#
# try() can keep counts or totals in global variables.  A loop can read
# successive sizes, usually with a reset of globals each time.
#
#     try(vpar) = total += vpar_diameter(vpar); 0;
#     for(n=8,13,
#        total=0;
#        read(Str("/tmp/trees-"n".gp"));
#        print(total));      \\ A001851 free trees total diameters
#
# The final return 0 in try() is a workaround for GP 2.9.x where return
# values of calls within read() accumulate on the stack.  This is noticeable
# when big n uses up a lot of stack unnecessarily.  Returning 0 avoids the
# problem (because 0 is a fixed special using no stack).
#
# The same sort of iteration can be done with vpar_free_next(), but
# externally generated trees saves that work.  The saving means fast enough
# to get up to around n=20, depending on what try()ed.
#
# The generated files become a bit big, for example around 124 mbytes for
# the roughly 2 million trees of n=21.
#

use 5.005;
use strict;
use FindBin;
use File::Which ();
use File::Temp;
use IPC::Run;

# uncomment this to run the ### lines
use Smart::Comments;


my @gentreeg_options = (
                        # '-Z3:5'   # diameter range
                       );
my $max_n = 21;

my $gentreeg_program = 'gentreeg';
if (! defined(File::Which::which($gentreeg_program))
    && defined(File::Which::which('nauty-gentreeg'))) {
  $gentreeg_program = 'nauty-gentreeg';
}

foreach my $num_vertices (1.. $max_n) {
  my $temp_fh = File::Temp->new;
  IPC::Run::run([ $gentreeg_program,
                  '-p', @gentreeg_options, $num_vertices ],
                '>',$temp_fh);

  seek $temp_fh, 0, 0 or die;
  open my $out, '>', "/tmp/trees-$num_vertices.gp" or die;
  my $count = 0;
  while (defined(my $str = readline $temp_fh)) {
    $str =~ tr/ /,/;
    $str =~ s/^.*$/try([$&]);/mg;
    print $out $str;
    $count++;
  }
  print "$count lines\n";
  close $out or die;
}
exit 0;

__END__


# IPC::Run::run(['nauty-gentreeg','-p',$num_vertices],
#               '>',"/tmp/trees-$num_vertices.vpar");
# open my $in, '<', "/tmp/trees-$num_vertices.vpar";
# open my $out, '>', "/tmp/trees-$num_vertices.gp" or die;

{
  # vpar free trees by gentreeg for use with vpar.gp

  foreach my $num_vertices (14) {
    require IPC::Run;
    print "run\n";
    my $h = IPC::Run::start([$gentreeg_program,'-p',$num_vertices],
                            '<', '/dev/null',
                            '>pipe',\*PIPE,
                            debug => 1);
    print "running\n";
    print "fileno ",fileno(PIPE),"\n";
    my $count = 0;
    for (;;) {
      print "readline\n";
      my $str = readline PIPE;
      if (!defined $str) { last; }
      print $str;
      $count++;
    }
    $h->finish;
    print "$count lines\n";
  }
  exit 0;
}
