\\ Copyright 2019, 2020, 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <binomial-tree-W.gp
\\
\\ This is some code calculating the Wiener index of a binomial tree of n
\\ vertices (vpar_make_binomial_tree()), and some related measures.
\\
\\ A tree of "order" k is n=2^k vertices and has a straightforward
\\ recurrence per
\\
\\     K. Viswanathan Iyer and K. R. Udaya Kumar Reddy, "Wiener index of
\\     Binomial Trees and Fibonacci Trees", Intl J Math Engg with Comp, 2009,
\\     http://arxiv.org/abs/0910.4432
\\
\\ They go recursively by blocks.  An arbitrary n can go by blocks too, but
\\ with attention to the relative depths (and sizes).  The build-up here
\\ through Wtop and Wend is slightly lengthy, but shows a couple of
\\ approaches to each.
\\


\\-----------------------------------------------------------------------------
\\ Wtop

\\ Let Wtop(n) be the total path lengths from each vertex up to the root,
\\ so total depths.
\\
Wtop(n) = vecsum(vpar_depths_vector(vpar_make_binomial_tree(n)));

{
  print1("Wtop(n)    ");
  for(n=0,16, print1(Wtop(n),","));
  print("     (A000788)");
}

\\ Depth is the number of 1-bits of v so sum of hammingweight().
vector(100,n,n--; Wtop(n) ) == \
vector(100,n,n--; sum(v=0,n-1, hammingweight(v)) )     || error();

\\ This sum is OEIS A000788, though that sequence is sum over 0..n,
\\ so n+1 terms, whereas here prefer 0..n-1 so n terms.
A000788(n) = Wtop(n+1);

my(A000788_samples=[0,1,2,4,5,7,9,12,13,15,17,20,22,25,28,32,33,35,37]); \
vector(#A000788_samples,n,n--; A000788(n)) == A000788_samples  || error();

\\ A000788 has many references and formulas.  Case Wtop(2^k) is
WtopK(k) = k/2 * 2^k;
{
  print1("Wtop(2^k)  ");
  for(k=0,9, print1(WtopK(k),","));
  print("               (A001787)");
}
vector(8,k,k--; Wtop(2^k) ) == \
vector(8,k,k--; k/2 * 2^k)   || error();

A001787(n) = WtopK(n);
my(A001787_samples=[0,1,4,12,32,80,192,448,1024,2304,5120,11264,24576,53248]); \
vector(#A001787_samples,n,n--; A001787(n)) == A001787_samples   || error();

\\ WtopK is total number of 1-bits among all the k-bit numbers v=0 to 2^k-1
\\ inclusive.  All combinations of 0 or 1 occur, so mean k/2 per vertex and
\\ hence k/2 * 2^k.
\\
\\ Or, v and its 0<->1 complement can be paired up
\\
\\        v = 1001110    depth 4
\\        v = 0110001    depth 3
\\            -------
\\            1111111    total 7
\\
\\ which is exactly k many 1-bits per pair, hence mean k/2 each in pairs.
\\ For k=0, there is just 1 vertex v=0, but the formula holds there too.


\\ A doubling of Wtop follows by noting that binomial tree 2*n is a new leaf
\\ vertex added underneath every vertex of tree n.  So additional Wtop(n)
\\ down to the original vertices, and n more for each of them 1 deeper.
\\ (Ralf Stephan formula in A000788.)
\\
vector(100,n,n--; Wtop(2*n) ) == \
vector(100,n,n--; 2*Wtop(n) + n)      || error();

\\ Going to 2*n-1 is 1 fewer new leaf, so only additional Wtop(n-1) down to
\\ the new leaves.  (Again Ralf Stephan formula in A000788.)
\\
vector(100,n, Wtop(2*n-1) ) == \
vector(100,n, Wtop(n) + Wtop(n-1) + n-1)    || error();
\\
\\ This is not particularly helpful for computer calculation since it splits
\\ to recurse into both Wtop(n) and Wtop(n-1).  Easier is 2*n+1 by adding
\\ the new vertex depth.
\\
vector(100,n, Wtop(2*n+1) ) == \
vector(100,n, Wtop(2*n) + hammingweight(2*n))    || error();

\\ A bit-wise form for Wtop follows in a similar way to the bit recurrence
\\ given by Shreevatsa R in A000788.  In the Wtop here, each bit of n
\\ corresponds to a 2^k sub-tree.  For example
\\
\\     0
\\     | \  \      \                 binomial tree
\\     1  2   4      8               n = 13 vertices
\\        |   | \    | \    \          = 1011 binary
\\        3   5  6   9  10   12        = 2^3 + 2^2 + 2^0
\\               |      |
\\               7      11
\\
\\     \---------/   \---/  \-/
\\        2^3         2^2   2^0
\\
\\ Total depths in the 2^3 block are WtopK(3).
\\ Total depths in the 2^2 block are WtopK(2) plus each is 1 deeper.
\\ Total depths in the 2^0 block are WtopK(0) plus each is 2 deeper.
\\
\\ The additional depth is how many 1-bits above the 2^2 or 2^0 etc
\\ position.  So write descending powers
\\
\\     n = 2^k[0] + 2^k[1] + 2^k[2] + ...
\\     where k[0] > k[1] > k[2] > ...
\\
\\     Wtop(n) = sum   WtopK(k[i]) + i*2^k[i]
\\               i>=0
\\
\\             = sum  (k[i]/2 + i) * 2^k[i]

\\ Return a vector of the "k" terms of n, from highest to lowest.
\\ so return vector k with n = 2^k[1] + 2^k[2] + ...
binary_powers_hightolow(n) =
{
  n>=0 || error();
  my(b=binary(n));
  my(vec=select(bit->bit!=0, b, 1));  \\ Vecsmall of indices
  vector(#vec,i, #b-vec[i]);
}
{
  for(n=0,100,
     my(k=binary_powers_hightolow(n));
     k == vecsort(k,,4+8)     || error();   \\ descending and distinct
     sum(i=1,#k, 2^k[i]) == n || error());  \\ sum to n
}

Wtop_by_powers(n) =
{
  my(k=binary_powers_hightolow(n));
  sum(i=1,#k,  (k[i]/2 + (i-1))*2^k[i]);
}
vector(100,n,n--; Wtop(n)) == \
vector(100,n,n--; Wtop_by_powers(n))    || error();

\\ This can be thought of as putting a "digit" value k[i]/2 + i at
\\ respective k[i] bit positions.
\\
\\       2^3     2^2     2^0     = 13 = n
\\      3/2+0   2/2+1   0/2+2    = 22 = Wtop(13)

Wtop(13) == 22  || error();
(3/2+0)*2^3 + (2/2+1)*2^2 + (0/2+2)*2^0 == Wtop(13)  || error();

\\ In GP up to version 2.13, fromdigits() doesn't like fractions 1/2 in its
\\ "digits" and a couple of 2.9.x versions didn't like anything except 0,1.
\\ Pol() and subst() can evaluate arbitrary powers.
\\
\\ fromdigits_any(0 is like fromdigits(), but arbitrary "digit" values in
\\ vec.  (But not polynomials in 'x.)
fromdigits_any(vec,base) = subst(Pol(vec),'x,base);

Wtop_by_binary(n) =
{
  my(vec=binary(n),            \\ bits high to low
     i=0);                     \\ count of 1-bits
  for(pos=1,#vec,
     if(vec[pos],              \\ at a 1-bit
        my(k=#vec-pos);
        vec[pos] = k/2 + i;    \\ new "digit" value
        i++));
  fromdigits_any(vec,2);
}
vector(100,n,n--; Wtop(n)) == \
vector(100,n,n--; Wtop_by_binary(n))    || error();

\\ Or bittest() and sum rather than building any vectors.
\\
Wtop_by_bittest(n) =
{
  my(ret=0, i=0);
  forstep(k=logint(n+1,2),0,-1,  \\ bit position high to low
     if(bittest(n,k),
        ret += (k/2 + i) << k;
        i++));
  ret;
}
vector(100,n,n--; Wtop(n)) == \
vector(100,n,n--; Wtop_by_bittest(n))    || error();


\\ Ralf Stephan in A000788 gives generating function.
gA000788(k_limit) = 1/(1-x)^2 * sum(k=0,k_limit, x^(2^k)/(1+x^(2^k)));
gA000788(8) + O(x^256) == sum(n=0,255, A000788(n)*x^n)   || error();

\\ This follows from the bits of n.
\\
\\ Firstly, a generating function for terms at 2^k .. 2*2^k-1 is a usual sum
\\
\\                                          x^(2^k) - x^(2^(k+1))
\\     x^(2^k) + ... + x^(2^(k+1) - 1)  =   ---------------------
\\                                                 (1-x)
\\
\\ Repeating such a run in each x^(2^(k+1)) block is all n with a 1-bit at
\\ bit position k.
\\
\\      x^(2^k) - x^(2^(k+1))
\\     -----------------------
\\     (1-x) (1 - x^(2^(k+1)))
\\
\\              x^(2^k)             by cancelling factor (1 - x^(2^k))
\\     =  -------------------       in numerator and denominator
\\        (1-x) (1 + x^(2^k))

\\ Generating function with 1*x^n at each n which has a 1-bit at bit
\\ position k, where k=0 is the lowest bit.
gbit(k) = x^(2^k) / (1-x) / (1 + x^(2^k));
for(k=0,7, gbit(k) + O(x^256) == sum(n=0,255, bittest(n,k)*x^n)  || error());

\\ Sum over all bit positions k is count 1-bits of n (A000120),
ghammingweight(k_limit) = sum(k=0,k_limit, gbit(k));
ghammingweight(8) + O(x^256) == sum(n=0,255, hammingweight(n)*x^n) || error();

\\ Wtop which is A000788 is then cumulative such count.
gA000788(8) == 1/(1-x) * ghammingweight(8)   || error();

gWtop(k_limit) = x*gA000788(8);
gWtop(8) + O(x^256) == sum(n=0,255, Wtop(n)*x^n)   || error();


\\-----------------------------------------------------------------------------
\\ Wtop Separate

\\ The "k/2" and "i" parts in Wtop can be taken separately.
\\
\\ k/2*2^k is a kind of "positional powers" sum half the position k at each
\\ 1-bit position k.
PosPowers(n) =
{
  sum(k=0,logint(n+1,2),
      if(bittest(n,k), k/2 * 2^k));
}
{
  print1("PosPowers    ");
  for(n=0,16, print1(PosPowers(n),","));
  print("   (A136013)");
}

\\ PosPowers is A136013, per the sum form by Marc LeBrun there.
A136013(n) = PosPowers(n);
my(A136013_samples=[0,0,1,1,4,4,5,5,12,12,13,13,16,16,17,17,32,32,33,33]); \
vector(#A136013_samples,n,n--; A136013(n)) == A136013_samples   || error();


\\ The "i" part sum of i*2^k in Wtop() is
Wtop_i_part(n) =
{
  my(ret=0, i=0);
  forstep(k=logint(n+1,2),0,-1,  \\ bit position high to low
     if(bittest(n,k),
        ret += i << k;
        i++));                   \\ how many 1s above, excluding self
  ret;
}
{
  print1("Wtop_i_part  ");
  for(n=0,16, print1(Wtop_i_part(n),","));
  print("           (A083741)");
}

\\ This is A083741 a(n) = l(n) + a(l(n)), where l(n) = n - 2^floor(log2(n)),
\\ l(n) = A053645(n) = n sans its highest 1-bit.
A083741(n) = Wtop_i_part(n);
my(A083741_samples=[0,0,0,1,0,1,2,4,0,1,2,4,4,6,8,11,0,1,2,4,4,6,8,11,8,10]); \
vector(#A083741_samples,n,n--; A083741(n)) == A083741_samples || error();

vector(100,n,n--; Wtop(n)) == \
vector(100,n,n--; PosPowers(n) + Wtop_i_part(n))   || error();

\\ "i" means bits of n are repeated in a sum, with i=0 times for the highest
\\ 1-bit, then i=1 for the second highest, etc, downwards.
\\
\\ Some bit-twiddling can locate the lowest 1-bit, and can use that to build
\\ up bits from low to high.  The low ones are repeated for each new 1-bit
\\ and hence get multiple i copies in the sum.
\\
\\ This sort of thing is best suited to a single machine word, since it
\\ crunches through modifying n, whereas say for a bignum just some
\\ bittest() on the input suffices.  It can be noted that on reaching n=0,
\\ the loop gets b=0 and n remaining 0.  So say 5 instructions in the loop
\\ could be repeated a few times pipelined CPUs where a block of
\\ instructions goes faster than looping.
\\
lowest_1_bit(n) = bitand(n,-n);   \\ lowest 1-bit of n = A006519
Wtop_i_part_by_bit_twiddling(n) =
{
  my(ret=0, low=0);
  while(n,
        my(b=lowest_1_bit(n));
        ret += low;               \\ repeated copies of the lows
        low += b;                 \\ each 1-bit successively
        n -= b);
  ret;
}
vector(100,n,n--; Wtop_i_part_by_bit_twiddling(n)) == \
vector(100,n,n--; Wtop_i_part(n))    || error();


\\ Also related in the OEIS is A298043 which is the same sort of descending
\\ "i" as A083741 = Wtop_i_part, but going from i=1 rather than i=0 at the
\\ highest 1-bit.  The effect is +1 at each bit position, and that is simply
\\ add n.
A298043(n) =
{
  my(ret=0, i=0);
  forstep(k=logint(n+1,2),0,-1,  \\ bit position high to low
     if(bittest(n,k),
        i++;                     \\ how many 1s above, including self
        ret += i << k));
  ret;
}
my(A298043_samples=[0,1,2,4,4,6,8,11,8,10,12,15,16,19,22,26,16,18,20,23]); \
vector(#A298043_samples,n,n--; A298043(n)) == A298043_samples || error();

vector(100,n,n--; A298043(n)) == \
vector(100,n,n--; A083741(n) + n)    || error();


\\ A generating function for PosPowers() follows from gbit above to
\\ identify the 1-bits which are to get factor k/2,
\\
gPosPowers(k_limit) = sum(k=0,k_limit, k/2*2^k * gbit(k));

gPosPowers(8) + O(x^256) == sum(n=0,255, PosPowers(n)*x^n)   || error();

\\ A generating function for Wtop_i_part() is difference Wtop - PosPowers.
gWtop_i_part(k_limit) = sum(k=0,k_limit, x/(1-x)*gbit(k) - k/2*2^k*gbit(k));
gWtop_i_part(k_limit) = sum(k=0,k_limit, ( x/(1-x) - k/2*2^k )*gbit(k));

gWtop_i_part(8) + O(x^256) == sum(n=0,255, Wtop_i_part(n)*x^n)   || error();

print();


\\-----------------------------------------------------------------------------
\\ Wend

\\ Let Wend(n) be the total path lengths from vertex n to each of its
\\ preceding vertices v = 0 .. n-1.
\\
\\     0              4 to 0   distance 1      1
\\     | \  \         4 to 1,2 distance 2    2*2
\\     1  2   4       4 to 3   distance 3      3
\\        |                                  -----
\\        3                                    8  = Wend(4)
\\

\\ Such path lengths are calculated by vpar_path_lengths_vector(), and here
\\ in a binomial tree of n+1 vertices.  Vertex numbers are 1 .. n+1,
\\ representing 0..n, so paths are to the last vertex n+1.

Wend(n) = vecsum(vpar_path_lengths_vector(vpar_make_binomial_tree(n+1),n+1));
{
  print1("Wend(n)      ");
  for(n=0,15, print1(Wend(n),","));
  print();
}
Wend(4) == 8   || error();

\\ A bit-wise calculation can be made in a similar way to Wtop.
\\ For example,
\\
\\     0
\\     | \  \      \                 binomial tree
\\     1  2   4      8               n = 13 last vertex
\\        |   | \    | \    \          = 1011 binary
\\        3   5  6   9  10   12        = 2^3 + 2^2 + 2^0
\\               |      |       \
\\               7      11       13
\\
\\     \---------/   \---/  \-/
\\        2^3         2^2   2^0
\\
\\ Total distance 2^3 block is its depths WtopK(3) plus 3 each up from 13.
\\ Total distance 2^2 block is its depths WtopK(2) plus 2 each up from 13.
\\ Total distance 2^0 block is its depths WtopK(0) plus 1 each up from 13.
\\
\\ Each block from smallest to biggest is 1 depth level higher above n.
\\ So write ascending powers
\\
\\     n = 2^k[1] + 2^k[2] + 2^k[3] + ...
\\     where k[1] < k[2] < k[3] < ...
\\
\\     Wend(n) = sum  WtopK(k[i]) + i*2^k[i]
\\               i>=1
\\
\\             = sum  (k[i]/2 + i) * 2^k[i]
\\

binary_powers_lowtohigh(n) = Vecrev(binary_powers_hightolow(n));
{
  for(n=0,100,
     my(k=binary_powers_lowtohigh(n));
     k == vecsort(k,,8)       || error();   \\ ascending and distinct
     sum(i=1,#k, 2^k[i]) == n || error());  \\ sum to n
}

Wend_by_powers(n) =
{
  my(k=binary_powers_lowtohigh(n));
  sum(i=1,#k,  (k[i]/2 + i)*2^k[i]);
}
vector(100,n,n--; Wend(n)) == \
vector(100,n,n--; Wend_by_powers(n))    || error();

\\ Or bittest() and sum rather than building any vectors.
\\
Wend_by_bittest(n) =
{
  my(ret=0, i=1);
  for(k=0,logint(n+1,2),  \\ bit position low to high
     if(bittest(n,k),
        ret += (k/2 + i) << k;
        i++));
  ret;
}
vector(100,n,n--; Wend(n)) == \
vector(100,n,n--; Wend_by_bittest(n))    || error();


\\ The "k/2" terms in Wend are the same PosPowers() as are in Wtop().
\\ The "i" terms of Wend differ from Wtop in counting how many 1-bits above
\\ instead of below.
\\
Wend_i_part(n) =
{
  my(ret=0, i=1);
  for(k=0,logint(n+1,2),  \\ bit position low to high
     if(bittest(n,k),
        ret += i << k;
        i++));            \\ how many 1s below, including self
  ret;
}
{
  print1("Wend_i_part  ");
  for(n=0,16, print1(Wend_i_part(n),","));
  print("  (A298011)");
}

vector(100,n,n--; Wend(n)) == \
vector(100,n,n--; PosPowers(n) + Wend_i_part(n))   || error();

\\ Wend_i_part() can use the same bit-twiddling as
\\ Wtop_i_part_by_bit_twiddling() to locate lowest 1-bit.
\\ Wend_i_part() is high bits repeats, so it is sum with repeatedly peeling
\\ off the lowest 1-bit (iterations of A129760).
\\
Wend_i_part_by_bit_twiddling(n) =
{
  my(ret=0);
  while(n, ret+=n; n-=lowest_1_bit(n));
  ret;
}
vector(100,n,n--; Wend_i_part(n)) == \
vector(100,n,n--; Wend_i_part_by_bit_twiddling(n))  || error();


\\ Wend_i_part() is the sum of i at bit positions made by OEIS A298011.
A298011(n) = Wend_i_part(n);
my(A298011_samples=[0,1,2,5,4,9,10,17,8,17,18,29,20,33,34,49,16,33,34,53]); \
vector(#A298011_samples,n,n--; A298011(n)) == A298011_samples || error();

\\ The recurrences given in A298011 (Remy Sigrist and David A. Corneth) are
\\ as follows.  In both cases 2* shifts the bits upwards.  For the odd case,
\\ adding 2*n is an extra copy of that shifted n so as to increase the "i"
\\ of those bits above the new low 1-bit.
\\
vector(100,n,n--; A298011(2*n)) == \
vector(100,n,n--; 2*A298011(n))  || error();
vector(100,n,n--; A298011(2*n+1)) == \
vector(100,n,n--; 2*A298011(n) + 2*n + 1)  || error();

\\ The generating function relation by Robert Israel expresses these
\\ recurrences.
my(g=sum(n=0,255, Wend_i_part(n)*x^n) + O(x^256)); \
g == 2*(x+1)*subst(g,'x,'x^2) + x*(1+x^2)/(1-x^2)^2   || error();

\\ The added term x*(1+x^2)/(1-x^2)^2 is n*x^n at odd n.
\\ It follows from some usual generating function forms.
\\ (Or such 0,odd,0,odd is a linear recurrence so my recurrence_guess() can
\\ go straight to it.)
\\
\\ All 1 coefficients,
gOnes(x) = 1/(1-x);
gOnes(x) + O(x^256) == sum(n=0,255, 1*x^n)  || error();
\\
\\ Partial sums are n at x^n (or the same by differentiating),
gN(x) = x/(1-x) * gOnes(x);
gN(x) + O(x^256) == sum(n=0,255, n*x^n)  || error();
\\
\\ Select odd terms by g(x)-g(-x).
gOddN(x) = 1/2*(gN(x) - gN(-x));
gOddN(x) + O(x^256) == sum(n=0,255, if(n%2==1, n*x^n))  || error();
gOddN(x) == x*(1+x^2)/(1-x^2)^2  || error();

\\ Expanding the g.f. relation and puts factor (1+x) and substituted x^2.
\\ Doing so k times is a term
\\
\\     gOddN(x^(2^k)) * (1+x)*(1+x^2)*(1+x^4)*...**(1+x^(2^(k-1)))
\\
\\ The product is repeated spreads so gives a run of terms
\\
\\     1 + x + x^2 + ... + x^(2^k-1)    = (1 - x^(2^k)) / (1-x)
\\
\\ and therefore,
\\
gWend_i_part(k_limit) = \
  sum(k=0,k_limit, 2^k * gOddN(x^(2^k)) * (1 - x^(2^k))/(1-x) );
gWend_i_part(8) + O(x^256) == sum(n=0,255, Wend_i_part(n)*x^n)   || error();

\\ The direct interpretation of each term is that at each 1-bit, add all the
\\ bits at and above.  Such additions give the target "i" copies.
\\ The 1+x+x^2+...+x^(2^k-1) is the bits below k having any of 2^k many
\\ values so blocks of n (repeated blocks).

\\ Some cancelling between numerator and denominator gives
gWend_i_part(k_limit) = \
  sum(k=0,k_limit, 2^k * x^(2^k)*(-1 + 2/(1 - x^(2^(k+1)))) /(1-x) /(1 + x^(2^k)) );
gWend_i_part(8) + O(x^256) == sum(n=0,255, Wend_i_part(n)*x^n)   || error();

\\ The denominators are close to gbit(k) and some rearranging expresses with
\\ that factor, though not in a particularly enlightening way.
gWend_i_part(k_limit) = \
  sum(k=0,k_limit, 2^k * (-1 + 2/(1-x^2^(k+1)) ) *gbit(k) );
gWend_i_part(8) + O(x^256) == sum(n=0,255, Wend_i_part(n)*x^n)   || error();


\\ Wend_i_part() can also be reached from Wtop_i_part().  The latter counts
\\ how many 1-bits below, and Wend_i_part() counts how many 1-bits above.
\\ So their total is hammingweight() at every bit, which means
\\ n*hammingweight(n).
\\
vector(100,n,n--; Wtop_i_part(n) + Wend_i_part(n)) == \
vector(100,n,n--; n*hammingweight(n))  || error();
\\
A245788(n) = n*hammingweight(n);
my(A245788_samples=[0,1,2,6,4,10,12,21,8,18,20,33,24,39,42,60,16,34,36,57]); \
vector(#A245788_samples,n,n--; A245788(n)) == A245788_samples || error();

\\ In ghammingweight(), multiplying n is by a usual derivative and x* shift
\\ up, so the two i_part's sum in terms of ghammingweight() is
\\
gWtop_i_part(8) + gWend_i_part(8) == x*ghammingweight(8)' + O(x^256) \
   || error();

\\ Difference gWtop_i_part = x*ghammingweight' - gWtop_i_part and working
\\ through some slightly tedious simplifications of its gbit()s leads to a
\\ different type of sum for gWtop_i_part,
\\
gWend_i_part(k_limit) = \
  sum(k=0,k_limit, 2^k*( k/2 + 1/(1+x^(2^k)) )*gbit(k) );
gWend_i_part(8) + O(x^256) == sum(n=0,255, Wend_i_part(n)*x^n)   || error();

\\ The bit-wise interpretation of this is to start from factor i = 1+k/2.
\\ The actual desired i should range 1+0 to 1+k according to how many 1-bits
\\ below.  The k/2 start imagines an "average" 1/2 1-bits and 1/2 0-bits.
\\ Then at each 1-bit below, add +1/2 to all above and at each 0-bit add
\\ -1/2 to all above, thereby increasing or decreasing to the actual i.
\\
\\ The simplest is k=0, where the desired addition is a floor(n/2) at 2*,
\\              0   1   2   3   4   5
\\     1/2* (  -0, +0, -1, +1, -2, +2 )  *2 up
\\
gNpm(x) = (-1+x)*gN(x^2);
gNpm(x) == x^2*(-1+x) / (1-x^2)^2    || error();
gNpm(x) + O(x^256) == sum(n=0,255, -(-1)^n * floor(n/2) * x^n)  || error();

\\ Then spread and repeat in a similar way to gOddN above.
gWend_i_part(k_limit) =
{
  sum(k=0,k_limit,
      2^k*( (1+k/2)*gbit(k) + gNpm(x^(2^k))*(1 - x^(2^k))/(1-x) ) );
}
gWend_i_part(8) + O(x^256) == sum(n=0,255, Wend_i_part(n)*x^n)  || error();
\\
\\ Expanding gNpm(x^(2^k)) and writing the result instead in terms of gbit
\\ leads to the ( k/2 + 1/(1+x^(2^k)) )*gbit(k) form above.

\\ The whole of gWend() has gPosPowers() which is another k/2, hence
gWend(k_limit) = sum(k=0,k_limit, 2^k*( k + 1/(1+x^(2^k)) )*gbit(k) );
gWend(8) + O(x^256) == sum(n=0,255, Wend(n)*x^n)  || error();

print();


\\-----------------------------------------------------------------------------
\\ Wiener Index

W(n) = vpar_Wiener_index(vpar_make_binomial_tree(n));
{
  print1("W(n)       ");
  for(n=0,14, print1(W(n),","));
  print();
}

\\ For a binomial tree of n=2^k many vertices ("order" k), the Wiener index
\\ is calculated by
\\
\\     K. Viswanathan Iyer and K. R. Udaya Kumar Reddy, "Wiener index of
\\     Binomial Trees and Fibonacci Trees", Intl J Math Engg with Comp, 2009,
\\     http://arxiv.org/abs/0910.4432
\\
WK(k) = (k-1)/2*4^k + 1/2*2^k;
{
  print1("W(2^k)     ");
  for(k=0,6, print1(WK(k),","));
  print("                         (A192021)");
}
vector(8,k,k--; WK(k)) == \
vector(8,k,k--; W(2^k))  || error();

\\ This is OEIS A192021.
A192021(n) = WK(n);  /* OFFSET=0, first value 0 */
\\
my(A192021_samples=[0, 1, 10, 68, 392, 2064, 10272, 49216, 229504, 1048832]); \
vector(#A192021_samples,n,n--; A192021(n)) == A192021_samples || error();

vector(100,n,n--; A192021(n)) ==  /* Colin Barker, amended for initial 0 */  \
vector(100,n,n--; 2^(n-1)*(1+2^(n)*(n-1)))  || error();
vector(100,n,n+=2; A192021(n)) ==  /* Colin Barker */  \
vector(100,n,n+=2; 10*A192021(n-1) - 32*A192021(n-2) + 32*A192021(n-3))  || error();
gA192021(x) = x/((1-2*x)*(1-4*x)^2);  /* Colin Barker, amended for initial 0 */
gA192021(x) + O(x^100) == sum(n=0,99, A192021(n)*x^n)  || error();

\\ Iyer and Reddy calculate this top-down from noting that tree order k is
\\ two copies of order k-1 connected by an edge between their roots.
\\
\\ So WK(k-1) in each half, and between halves WtopK(k-1) to the root, times
\\ 2^(k-1) vertices in the other half, the same in the other half, and the
\\ connecting edge crossed (2^(k-1))^2 times.
\\
vector(100,k, WK(k)) == \
vector(100,k, 2*WK(k-1) + 2*WtopK(k-1)*2^(k-1) + (2^(k-1))^2 )  || error();


\\ For W(n) of any number of vertices, the same top-down can be applied to
\\ the blocks as for Wtop.
\\
\\ Paths within a block are WK.
\\
\\ Paths between a pair of blocks are WtopK(k1) to its top, times 2^k2
\\ vertices in the other, then vice versa, and distance to the other.  Since
\\ WtopK has factor 2^k, each of these components becomes factor 2^(k1+k2).
\\ So with n in powers of 2 again,
\\
\\     n = 2^k[0] + 2^k[1] + 2^k[2] + ...
\\     where k[0] > k[1] > k[2] > ...
\\
\\                                  /  WtopK(k[i])*2^k[j]
\\     W(n) = sum   WK(k[i]) + sum  |  + WtopK(k[i])*2^k[j]
\\            i>=0             j<i  \  + (j-i)*2^k[i]*2^k[j]
\\
\\          = sum   WK(k[i]) + sum  (k[i]/2 + k[j]/2 + i-j) * 2^(k[i]+k[j])
\\            i>=0             j<i
\\
\\ The resulting sum is symmetric in i,j so the powers k[] can be taken
\\ either high to low or low to high.  Only their distance apart i-j is used
\\ (whichever way around is positive).
\\
\\ WK part 1/2*2^k[i] sums to simply n/2, being 1/2 at each 1-bit of n, so
\\ that can be taken out of the sum.
\\
\\ WK part (k-1)/2*4^k is almost but not quite the same as an i=j in the
\\ double-sum.
\\
\\     (k/2 + k/2 + 0)*2^(k + k) = k*4^k   != (k-1)/2*4^k
\\
\\ The double-sum of i=j would be as if path between a pair of vertices went
\\ always up to the root and back down, rather than only up to the deepest
\\ common ancestor.

W_by_powers(n) =
{
  my(k=binary_powers_hightolow(n));
  n/2 + sum(i=1,#k,
            (k[i]-1)/2*4^k[i]
            + sum(j=1,i-1,
                  (k[i]/2 + k[j]/2 + i-j)*2^(k[i] + k[j])));
}
vector(200,n,n--; W(n)) == \
vector(200,n,n--; W_by_powers(n))   || error();


\\ The bit-twiddling of Wend_i_part_by_bit_twiddling() can be used for the
\\ i-j part.
\\
\\ Let a=2^k[0], b=2^k[1], etc.  Then the i-j coefficient is products a*b,
\\ a*c, etc, with a repeat count which is the distance from the leading
\\ diagonal,
\\               i
\\          a  b  c  d  e
\\        a .  1  2  3  4       (a) * (b c d e)
\\     j  b    .  1  2  3       (a b) * (c d e)
\\        c       .  1  2       (a b c) * (d e)
\\        d          .  1       (a b c d) * (e)
\\        e             .
\\         i-j repeat count       products
\\
\\ These counts can be obtained by products of the form shown at the right.
\\ a*bcde + a*cde + a*de + a*e is the counts for the first row of the
\\ matrix.  b*cde + b*de + b*e is the second row, and so on.
\\ Notice this is not merely n^2 or similar.  An n^2 is count 1 everywhere.
\\
\\ The products are built by peeling off the lowest 1-bit each time, giving
\\ a "low" which is the right side of the above, and a "high" which is the
\\ left side.
\\
\\ The k[i]/2 and k[j]/2 parts are made by a product
\\
\\     (k[0]/2*2^k[0] + k[1]/2*2^k[1] + ... )
\\     *
\\     (       2^k[0] +        2^k[1] + ... )
\\
\\ The latter is simply n, so an n*PosPowers(n).  In the code below logint()
\\ gets the k from b.  (If going by bittest(n,k) then would have the k
\\ already.)  The initial p=1 gives the n/2 constant, after /2 at the end.
\\
\\ The cross terms k[i]*2^k[i]*2^k[i] = k[i]*4^k[i] are part outside the j
\\ sum in W_by_powers(), and the -1 there is by subtracting 4^k[i] which is
\\ b*b in the code below.
\\
W_by_bit_twiddling(n) =
{
  my(p=1,c=0, low=0,high=n);
  while(high,
        my(k=valuation(high,2), b=1<<k);
        low+=b; high-=b;
        c += 2*low*high - b^2;
        p += k*b);
  (n*p + c)/2;
}
vector(200,n,n--; W(n)) == \
vector(200,n,n--; W_by_bit_twiddling(n))    || error();


\\ Another approach for W is to go bottom-up by noting that tree 2*n is a
\\ new leaf underneath each existing vertex.  (Like bottom-up Wtop.)
\\
\\ So W(2*n) is 2*2*W(n) total distance between the pairs of original vertex
\\ and new leaf.  The edge to the new leaf is crossed by 2*n-1 paths to the
\\ other vertices, and there are n new leaves.
\\
vector(100,n,n--; W(2*n)) == \
vector(100,n,n--; 4*W(n) + n*(2*n-1))   || error();

\\ For n=2^(k-1) this is
\\
vector(100,k,k--; WK(k)) == \
vector(100,k,k--; 4*WK(k-1) + 2^(k-1)*(2*2^(k-1) - 1))   || error();

\\ Wend(n) is the W increment (in all cases).
\\ That increment and W(2*n) can be used to go to W(2*n+1) for a recurrence.
\\ Doing so by bits of n high to low has the effect of Wend examining bits
\\ in the triangular i,j style as in W_by_powers() (and which might be more
\\ work than the accumulated products in W_by_bit_twiddling()).
\\
vector(100,n,n--; W(n+1)) == \
vector(100,n,n--; W(n) + Wend(n))   || error();

\\ A straightforward generating function for W is cumulative Wend,
\\
gW(k_limit) = sum(k=0,k_limit, x/(1-x) * 2^k*( k + 1/(1+x^(2^k)) )*gbit(k) );
gW(8) + O(x^256) == sum(n=0,255, W(n)*x^n)  || error();


\\ The i-j part separately is
W_ij_part(n) =
{
  my(k=binary_powers_hightolow(n));
  sum(i=1,#k, + sum(j=1,i-1, (i-j)*2^(k[i] + k[j])));
}
{
  print1("W_ij_part  ");
  for(n=0,14, print1(W_ij_part(n),","));
  print();
}
\\ not in OEIS: 2,0,4,8,18,0,8,16,34,32,52,72

W_ij_part_by_bit_twiddling(n) =
{
  my(ret=0, low=0,high=n);
  while(high,
        my(b=bitand(high,-high));
        low+=b; high-=b;
        ret += low*high);
  ret;
}
vector(200,n,n--; W_ij_part(n)) == \
vector(200,n,n--; W_ij_part_by_bit_twiddling(n))    || error();

print();


\\-----------------------------------------------------------------------------
\\ Distances in Order k

\\ For binomial tree order k of 2^k vertices, Emeric Deutsch in OEIS A192020
\\ counts how many vertices are distance d apart, for d ranging 1 up to the
\\ diameter of the tree.

{
  print("Order k count of pairs by distance (A192020)");
  print("    dist  1    2    3    4    5    6    7    8    9");
  for(k=1,5,
     my(vpar=vpar_make_binomial_tree(2^k),
        m=vpar_path_lengths_matrix(vpar),
        counts=vector(max(0,2*k-1)));
     for(i=1,#vpar, for(j=i+1,#vpar, counts[m[i,j]]++));
     print1("  k="k"  ");
     for(d=1,#counts,
        printf("%4d%s", counts[d], if(d==#counts,"",",")));
     print();

     \\ distance 1 apart is across each edge
     counts[1] == vpar_num_edges(vpar)  || error();

     \\ diameter 2*k-1 is attained between a single pair
     counts[#counts] == 1                           || error();
     vpar_diameter(vpar) == #counts                 || error();
     vpar_diameter_and_count(vpar) == [#counts,1]   || error();
     );
}

\\ Emeric Deutsch makes a generating function for the counts from an r(k,x)
\\ which is number of vertices at depth in order k, and a recurrence for
\\ then a w(k,x) generating function for number of vertices at distance in
\\ order k.

\\ Polynomial count*x^d for count many vertices at depth d.
r(k,x) =
{
  my(vpar=vpar_make_binomial_tree(2^k),
     depths=vpar_depths_vector(vpar));
  sum(v=1,#vpar, x^depths[v]);
}

\\ This number of vertices at depth is binomial(k,d) -- which is why the
\\ tree is called the binomial tree.  The generating function for a row of
\\ binomials is simply (1+x)^k.
{
  for(k=0,7,
     r(k,x) == sum(d=0,k, binomial(k,d)*x^d)  || error();
     r(k,x) == (1+x)^k  || error();
  );
}

\\ Polynomial count*x^d for count many vertex pairs at distance d apart.
w(k,x) =
{
  my(vpar=vpar_make_binomial_tree(2^k),
     distances=vpar_path_lengths_matrix(vpar));
  sum(u=1,#vpar, sum(v=u+1,#vpar, x^distances[u,v]));
}

\\ The recurrence is two trees of k-1, plus paths between them by cross
\\ product of depths each vertex up to the root.  The second tree copy is 1
\\ depth deeper, so "x*" to add 1 to each such path.
{
  for(k=1,7,
     w(k,x) == 2*w(k-1,x) + x*r(k-1,x)^2  || error();
  );
}

\\ Expanding this recurrence is a sum.  The sum is rising and falling powers
\\ so becomes a rational function.
{
  for(k=0,7,
     w(k,x) == sum(i=0,k-1, 2^(k-1-i)*x*r(i,x)^2)     || error();
     w(k,x) == sum(i=0,k-1, 2^(k-1-i)*x*(x+1)^(2*i))  || error();
     w(k,x) == x*((x+1)^(2*k) - 2^k) / ((x+1)^2 - 2)  || error();  \\ t_RFRAC
  );
}
w(k,x) = x*((x+1)^(2*k) - 2^k) / ((x+1)^2 - 2);

\\ w(k,x) evaluated at x=1 is simply the number of vertex pairs.
\\ Derivative w(k,x)' is terms like d*count*x^(d-1) so has distance d times
\\ its count, so total is the Wiener index.
\\ An explicit formula for the derivative is straightforward, but not
\\ particularly suggestive of anything.  The form shown below is the close
\\ to WK(k) after substituting x=1 except denominator ((1+1)^2-2)^2 == 4 and
\\ the numerator is 4* bigger from various x factors.
{
  for(k=0,7,
     my(vpar=vpar_make_binomial_tree(2^k));
     w(k,1) == binomial(#vpar,2)     || error();
     my(wdash = w(k,x)');
     subst(wdash,x,1) == vpar_Wiener_index(vpar)  || error();

     wdash == ( (k*2*x*((x+1)^2-2) - (x+1)*(x^2+1)) * (x+1)^(2*k-1)
                 + 2*(x^2+1)                        * 2^(k-1)
              ) / ((x+1)^2-2)^2           || error();
     WK(k) == (k-1)*2^(2*k-1) + 2^(k-1)   || error();
  );
}


\\-----------------------------------------------------------------------------
print("end");
