\\ Copyright 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(recover,0)
default(strictargs,1);
read("../vpar.gp");


\\ Usage: gp <parking-to-labelled.gp
\\
\\ This is some code using the method of
\\
\\     Antonio Guedes de Oliveira, Michel las Vergnas, "Parking Functions
\\     and Labelled Trees", Seminaire Lotharingien de Combinatoire, volume
\\     65, 2011, article B65e
\\
\\ for a bijection between parking functions and labelled forests.
\\
\\ A "parking function" f[i] = pos is cars numbered i = 1,...,n each
\\ desiring to park at a position pos = 1,...,n.  Each car is parked
\\ successively.  If its preferred pos is already occupied then it parks at
\\ the next later place which is not occupied.
\\
\\ Only some f[i] vectors of preferred positions are parking functions.  If
\\ car i finds that other cars before it have left its preferred place
\\ occupied, and all other places through to n inclusive also occupied, then
\\ it cannot park and this set of places is not a parking function.  The
\\ simplest clash is two f[i]=f[j]=n since both i,j cannot both park at n
\\ and there are no positions at all after n for the second car.
\\
\\ If f is a permutation of 1..n then that is clearly a parking function
\\ since there's no duplication of desired places and all cars can park
\\ exactly at their preferred place.  For all parking functions, the actual
\\ places where cars end up parked is a permutation of 1..n (see P(f)
\\ below).


\\-----------------------------------------------------------------------------

\\ f is vector f[i] = pos desired to park for n=#f cars.
\\ Each pos must be 1 <= pos <= n.
\\ Return 1 if f is a parking function, or 0 if not.
\\
\\ A simple condition for being a parking function is that for a given pos
\\ there must be at least pos many f[i]<=pos, as per Oliveira and Vergnas
\\ definition 2.1.  If this is not so, then there are not enough cars
\\ wanting to park in 1..pos to fill those spaces, which will leave some
\\ empty there and an excess after.
\\
\\ It's probably more efficient to actually do the parking and check for not
\\ running off the end, though simple counting like this has no extra
\\ workspace and could run with some parallelism if desired.
\\
is_parking(f) =
{
  for(pos=1,#f,
     if(sum(j=1,#f, f[j]<=pos) >= pos || return(0)));
  1;
}

\\ The number of parking functions of length n is (n+1)^(n-1).
\\ This is also the number of labelled forests vpar_count(n,1)
\\ and hence the bijection Oliveira and Vergnas construct.
{
  for(n=1,5,
     my(count=0);
     forvec(f=vector(n,i, [1,n]),   \\ "autofunctions" all entries 1..n
           count += is_parking(f));
     print1(count", ");
     count == (n+1)^(n-1)      || error();
     count == vpar_count(n,1)  || error();
  );
  print(" ...    (A000272)");
  print();
}
  

\\ P(f) takes f a parking function vector, so f[i] = pos desired.
\\ Return a vector sigma[i] = actual position car i has parked.
\\ But "sigma" is a builtin function, so variously in the code use variable
\\ name "s".
\\
\\ Oliveira and Vergnas definition of P (in turn following Pollak), allows a
\\ position n+1 and wrap-around.  If n+1 is unoccupied then this is a
\\ parking function.  For the code here, this is not necessary so the return
\\ is just where 1..n park.
\\
\\ Each i parks at its own distinct position, so the return is a permutation.
\\ It is returned as a Vecsmall as usual for GP permutations, allowing
\\ perm^-1 for inverse, "*" for composition, etc.
\\
P(f) =
{
  my(ret=vectorsmall(#f),
     occupied=vectorsmall(#f));
  for(i=1,#f,
     my(pos=f[i]);          \\ preferred place
     while(occupied[pos],   \\ search for pos unoccupied
           pos++;
           if(pos>#f, error("oops, not a parking function")));
     ret[i]=pos;
     occupied[pos]=1);
  ret;
}

\\ A "probe" is when a desired pos is occupied and so must look at the next.
\\ Pr = total probes, being total over all cars of actual park - desired pos.
\\
\\ For this total, there's no need to calculate the actual places.  They are
\\ a permutation of 1..n so their total in Pr is sum(1..n) == triangular
\\ number, then subtract the total of f.
\\
triangular(n) = (n+1)*n/2;    \\ sum 1..n
triangular(0) == 0    || error();
triangular(1) == 1    || error();
triangular(2) == 1+2  || error();
Pr_total_probes(f) = triangular(#f) - vecsum(f);

{
  \\ check Pr_total_probes() == sum(sigma - f)
  for(n=1,5,
     my(count=0);
     forvec(f=vector(n,i, [1,n]),   \\ "autofunctions" all entries 1..n
           if(is_parking(f),
              my(Pr=Pr_total_probes(f),
                 s=P(f));
              Pr == sum(i=1,n, s[i]-f[i]) || error())));
}

\\ For a permutation,
\\ dp[i] = count how many i-1,i-2,... occur in perm before position i.
\\ perm[1 .. i-1] is the part before position i.
\\ If i-1 does not occur there, then dp=0.
\\ If i-1 does occur, then try i-2.  If i-2 does not then just dp=1 for i-1.
\\ And so on, counting back until an i-x not occurring.
\\
\\ For any parking function f, sigma=P(f) permutation and 
\\   dp(f)[i] = dp(sigma)[i] - (sigma[i] - f[i])
\\ so the permutation count rule on the sigma actual places, then subtract
\\ how far car i had to go forward due to its pos occupied.
\\ Always have dp(sigma)[i] > (sigma[i] - f[i]) since that many consecutive
\\ occupied positions are in sigma preceding i.
\\
dp_of_vector(f) =
{
  my(s=P(f));
  vector(#s,i,
         my(count=0,
            elem=s[i]);
         for(c=1,i-1,
            elem--;
            for(j=1,i-1,
               if(s[j]==elem, count++; next(2)));
            break);
         count - (s[i]-f[i]));
}

\\ dt(v) is what Oliveira and Vergnas call "depth" of a vertex v.
\\ This is not depth meaning how far down from the root but instead
\\ dt(v) = how many descendants k below v which are smaller k < v.
\\
\\ In an upwards labelled tree, dt(v) is all below v, except itself, so
\\ vpar_subtree_size() - 1.  For non-upwards labelling, dt is smaller.
\\
dt_of_vertex(vpar,v) =
{
  sum(k=1,v-1,                       \\ k<v
      vpar_is_ancestor(vpar,k,v));   \\ and below v, so v is ancestor
}
dt_of_vpar(vpar) = vector(#vpar,v, dt_of_vertex(vpar,v));

\\ desc(v) is the descendants counted by dt(v), ie. those k below and k<v.
\\ The return is a vector of those vertices, in ascending order.
\\
desc_of_vertex(vpar,v) = select(k->vpar_is_ancestor(vpar,k,v), [1..v-1]);


\\ inv(v) is the number of "inversions" of v in forest vpar.
\\ This is how many descendants k below v which are bigger k > v.
\\
\\ Oliveira and Vergnas call a tree with no inversions at all "decreasing",
\\ meaning vertex numbers decrease as go down the tree.  This is increasing
\\ vertex numbers as go up the tree so vpar_is_upwards().
\\ 
inv_of_vertex(vpar,v) =
{
  sum(k=v+1,#vpar,                   \\ k>v
      vpar_is_ancestor(vpar,k,v));   \\ and below v, so v is ancestor
}
inv_of_vpar(vpar) = vector(#vpar,v, inv_of_vertex(vpar,v));


\\ phi_perm_to_upwards_vpar(perm) takes a permutation and returns an upwards
\\ labelled forest in the manner of Oliveira and Vergnas phi(sigma) mapping
\\ permutation sigma to a forest (or rather there a tree with extra vertex
\\ n+1 above).  They show this is a one-to-one correspondence between
\\ permutations and upwards forests.
\\
\\ For vertex v, look at perm[v] and then search among p = v+1,...,n for
\\ perm[p] > perm[v].  Take p as the position where perm[p] is smallest
\\ among all the perm[p] > perm[v].  Then v gets parent p.
\\ If there are no perm[p]>perm[v] at all, including when v=n is the last
\\ vertex, then v is a root.
\\ By construction, this is vpar_is_upwards().
\\
phi_perm_to_upwards_vpar(perm) =
{
  my(vpar=vector(#perm));
  for(v=1,#perm,
     my(min_p=0);  \\ default to new root
     for(p=v+1,#perm,
        if(perm[p] > perm[v],                    \\ must bigger than perm[v]
           if(!min_p || perm[p] < perm[min_p],   \\ new smallest perm[p]
              min_p = p)));
     vpar[v]=min_p);
  vpar;
}

{
  \\ check phi_perm_to_upwards_vpar() and that all vpar produced are distinct
  for(n=1,6,
     my(seen=List([]));
     for(i=1,n!,
        my(s=numtoperm(n,i),
           vpar=phi_perm_to_upwards_vpar(s));
        vpar_validate(vpar);
        vpar_is_upwards(vpar) || error();
        listput(seen,vpar);

        \\ Lemma 3.3
        dp_of_vector(s) == dt_of_vpar(vpar)  || error();
     );
     #Set(seen) == #seen  || error();       \\ all distinct
     #seen == n!          || error();
     #seen == vpar_count_updown(n,1) || error(); \\ upwards forests
     );        
}


print("Example parking function");
f=[1,8,5,2,7,4,4,8,1];
print("f     = "f);

s=P(f);
print("sigma = ",Vec(s)," = P(f) actual parks, permutation");
s == Vecsmall([1,8,5,2,7,4,6,9,3])    || error();

a = s^-1;
print("alpha = "Vec(a)"  = sigma^-1");
a == Vecsmall([1,4,9,6,3,7,5,2,8])    || error();

{
print("
Figure 1(b) example phi(sigma), sans the n+1 = 10 above the top
         8          9
       / | \\        |
      2  5  7       4
           / \\      |
          3   6     1");
}
vpar_b = phi_perm_to_upwards_vpar(s);
print("vpar      = "vpar_b);
vpar_validate(vpar_b);
vpar_b == [4,8,7,9, 8,7,8,0,0]  || error();

seq = vpar_seq_postorder(vpar_b,1);
print("linear    = "Vec(seq)" = seq postorder(down=1)  = alpha");
seq == Vecsmall([1,4,9,6,3,7,5,2,8])  || error();
seq == a                              || error();

dp = dp_of_vector(s);
print("dp(sigma) = "dp"  perm consecutives preceding");

dt = dt_of_vpar(vpar_b);
print("dt(vpar)  = "dt"  num smaller descendents below each v");

print("dt(sigma) == dp(forest) per lemma 3.3");
dp == dt  || error();


dt_of_vertex(vpar_b,8) == 5    || error();
dt_of_vertex(vpar_b,9) == 2    || error();
apply(i->s[i], [6,3,7,5,2]) == [4,5,6,7,8]   || error();


\\ phi_parking_to_forest(f) takes parking function f and returns vpar forest.
\\ This is the final phi(f) of Oliveira and Vergnas.
\\
\\ The forest starts from permutation sigma = P(f) and the resulting upwards
\\ vpar = phi_perm_to_upwards_vpar(sigma).  Then some cyclic relabellings
\\ are applied which change it to an arbitrary forest.
\\
\\ pr = sigma[i]-f[i] = how many places car i had to skip due to its
\\ preferred pos being already occupied (possibly pr=0).  A relabelling
\\ cycle of length = pr + 1 is applied to the forest.  The desc_of_vertex()
\\ vertices below i are taken, together with i itself, in descending order
\\
\\     i, a[1], a[2], ..., a[dp]
\\
\\ The first pr+1 of them are taken and rotated
\\
\\     from    i,     a[1], a[2], ... a[pr-1],  a[pr]
\\     to      a[pr],  i,   a[1], ... a[pr-2], a[pr-1]
\\
\\ This has the effect of bringing the small a[pr] up to the top i, and
\\ hence no longer upwards labelled.  These cyclic relabellings are
\\ successively applied for i=1,...,n.
\\
\\ Because the relabellings only act on vertices under a given i, they do
\\ not upset anything for later i further up the tree.  So the relabellings
\\ can be either applied immediately, or accumulated in a perm and applied
\\ at the end.  The latter is probably more efficient.  The code does both
\\ ways as a consistency check.

\\ alist is a vector of i, a[1], a[2], ... a[pr]
\\ Return a permutation (Vecsmall) length n which is the rotate shown above.
alist_to_perm(alist,n) =
{
  my(perm=numtoperm(n,0));   \\ identity perm
  for(i=1,#alist,
     my(j=i-1); if(j==0, j=#alist);
     perm[alist[i]] = alist[j]);    \\ map preceding elem, with wrap-around
  perm;
}

phi_parking_to_forest(f) =
{
  my(s=P(f),
     vpar=phi_perm_to_upwards_vpar(s),
     total_perm = numtoperm(#f,0));  \\ initial identity perm
  for(i=1,#f,
     \\ desc_of_vertex() is ascending, Vecrev to be descending
     my(alist = concat(i, Vecrev(desc_of_vertex(vpar,i))),
        pr = s[i] - f[i],
        len = pr + 1);
     alist = alist[1 .. pr+1];
     my(g = alist_to_perm(alist,#vpar));
     \\ print("  i="i" alist "alist"  perm g="Vec(g));
     vpar = vpar_relabel_perm(vpar,g);
     total_perm = g*total_perm;
  );
  vpar_relabel_perm(phi_perm_to_upwards_vpar(s), total_perm)
    == vpar || error();
  vpar;
}

\\ Figure 2 rewriting process examples.
\\ These start from phi(sigma) which is no rewriting, just perm to upwards,
\\ and then successively applied changes which reach the example f parking
\\ function.
\\
        s == Vecsmall([1,8,5,2,7,4, 6, 9, 3]) || error();
phi_parking_to_forest([1,8,5,2,7,4, 6, 9, 3]) == [4,8,7,9,8,7,8,0,0] ||error();
phi_parking_to_forest([1,8,5,2,7,4, 6, 9, 2]) == [9,8,7,0,8,7,8,0,4] ||error();
phi_parking_to_forest([1,8,5,2,7,4, 6, 9, 1]) == [0,8,7,9,8,7,8,0,1] ||error();
phi_parking_to_forest([1,8,5,2,7,4, 5, 9, 1]) == [0,8,6,9,8,8,6,0,1] ||error();
phi_parking_to_forest([1,8,5,2,7,4, 4, 9, 1]) == [0,8,8,9,8,3,3,0,1] ||error();
phi_parking_to_forest([1,8,5,2,7,4, 4, 8, 1]) == [0,7,7,9,7,3,0,3,1] ||error();
                 f == [1,8,5,2,7,4, 4, 8, 1]  || error();

\\ 7 pos 4
\\ past 6 was 4, 3 was 5, set 6     7,6,3
\\ 8 pos 8
\\ past 2 was 8, set 9              8,2       but cycle 8,7 are biggest
\\ 9 pos 1
\\ past 1 was 1, 4 was 2, set 3     9,1,4


{
print("
Figure 1(a) example phi(f), sans the n+1 = 10 above the top
      1        7
      |      / | \\
      9     2  3  5
      |       / \\
      4      6   8");
}

vpar_a = phi_parking_to_forest(f);
print("vpar   = "vpar_a);
vpar_validate(vpar_a);
vpar_a == [0,7,7,9,7, 3,0,3,1]  || error();

\\ dt is the annotations shown in red in figure 1(a), here as a vector at
\\ each vertex label
dt = dt_of_vpar(vpar_a);
print("dt     = "dt"   num smaller descendents below each v");
dt == [0,0,0,0, 0,0,4,0,1]    || error();

inv = inv_of_vpar(vpar_a);
print("inv    = "inv"   num bigger descendants below each v");
inv == [2,0,2,0,0,0,1,0,0]    || error();

df = dp_of_vector(f);
print("dp(f)  = "df);
df == [0,0,0,1, 0,0,0,4,0]    || error();


print();
{
  \\ Check that phi_parking_to_forest() gives different vpar from each
  \\ possible parking function.

  \\ The number of parking functions = number of forests = (n+1)^(n-1) grows
  \\ rapidly with size n, so this sort of explicit check can only go to
  \\ small n.  If you have a fast computer then could try to n=7 or 8.

  print1("checking ... ");
  for(n=1,5,
     my(seen=List([]));
     forvec(f=vector(n,i, [1,n]),   \\ "autofunctions" all entries 1..n
           is_parking(f) || next;
           my(vpar=phi_parking_to_forest(f));
           vpar_validate(vpar);
           listput(seen,vpar));

     #Set(seen) == #seen       || error();   \\ all distinct
     #seen == vpar_count(n,1)  || error();   \\ is count labelled forests
     #seen == (n+1)^(n-1)      || error();
  );
  print("good");
}

\\-----------------------------------------------------------------------------
